# stdlib
import json
from typing import Dict
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import final

# relative
from ...... import serialize
from ......proto.grid.messages.user_messages_pb2 import (
    CreateUserMessage as CreateUserMessage_PB,
)
from ......proto.grid.messages.user_messages_pb2 import (
    DeleteUserMessage as DeleteUserMessage_PB,
)
from ......proto.grid.messages.user_messages_pb2 import (
    GetCandidatesMessage as GetCandidatesMessage_PB,
)
from ......proto.grid.messages.user_messages_pb2 import (
    GetCandidatesResponse as GetCandidatesResponse_PB,
)
from ......proto.grid.messages.user_messages_pb2 import (
    GetUserMessage as GetUserMessage_PB,
)
from ......proto.grid.messages.user_messages_pb2 import (
    GetUserResponse as GetUserResponse_PB,
)
from ......proto.grid.messages.user_messages_pb2 import (
    GetUsersMessage as GetUsersMessage_PB,
)
from ......proto.grid.messages.user_messages_pb2 import (
    GetUsersResponse as GetUsersResponse_PB,
)
from ......proto.grid.messages.user_messages_pb2 import (
    ProcessUserCandidateMessage as ProcessUserCandidateMessage_PB,
)
from ......proto.grid.messages.user_messages_pb2 import (
    SearchUsersMessage as SearchUsersMessage_PB,
)
from ......proto.grid.messages.user_messages_pb2 import (
    SearchUsersResponse as SearchUsersResponse_PB,
)
from ......proto.grid.messages.user_messages_pb2 import (
    UpdateUserMessage as UpdateUserMessage_PB,
)
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.deserialize import _deserialize
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address


@serializable()
@final
class CreateUserMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        name: str,
        email: str,
        password: str,
        reply_to: Address,
        role: Optional[str] = "",
        website: str = "",
        institution: str = "",
        daa_pdf: Optional[bytes] = b"",
        msg_id: Optional[UID] = None,
        budget: Optional[float] = 0.0,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.email = email
        self.password = password
        self.role = role
        self.name = name
        self.daa_pdf = daa_pdf
        self.website = website
        self.institution = institution
        self.budget = budget

    def _object2proto(self) -> CreateUserMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: CreateUserMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return CreateUserMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            email=self.email,
            password=self.password,
            role=self.role,
            name=self.name,
            institution=self.institution,
            website=self.website,
            daa_pdf=self.daa_pdf,
            reply_to=serialize(self.reply_to),
            budget=self.budget,
        )

    @staticmethod
    def _proto2object(
        proto: CreateUserMessage_PB,
    ) -> "CreateUserMessage":
        """Creates a CreateUserMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: CreateUserMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return CreateUserMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            email=proto.email,
            password=proto.password,
            name=proto.name,
            role=proto.role,
            website=proto.website,
            institution=proto.institution,
            daa_pdf=proto.daa_pdf,
            reply_to=_deserialize(blob=proto.reply_to),
            budget=proto.budget,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return CreateUserMessage_PB


@serializable()
@final
class GetUserMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        user_id: int,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.user_id = user_id

    def _object2proto(self) -> GetUserMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetUserMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetUserMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            user_id=self.user_id,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetUserMessage_PB,
    ) -> "GetUserMessage":
        """Creates a GetUserMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetUserMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetUserMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            user_id=proto.user_id,
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetUserMessage_PB


@serializable()
@final
class ProcessUserCandidateMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        candidate_id: int,
        status: str,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.candidate_id = candidate_id
        self.status = status

    def _object2proto(self) -> ProcessUserCandidateMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetUserMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return ProcessUserCandidateMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            candidate_id=self.candidate_id,
            status=self.status,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: ProcessUserCandidateMessage_PB,
    ) -> "ProcessUserCandidateMessage":
        """Creates a GetUserMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetUserMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return ProcessUserCandidateMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            candidate_id=proto.candidate_id,
            status=proto.status,
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return ProcessUserCandidateMessage_PB


@serializable()
@final
class GetUserResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.content = content

    def _object2proto(self) -> GetUserResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetUserResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=serialize(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: GetUserResponse_PB,
    ) -> "GetUserResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetUserResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=_deserialize(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetUserResponse_PB


@serializable()
@final
class GetUsersMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)

    def _object2proto(self) -> GetUsersMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetUsersMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetUsersMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetUsersMessage_PB,
    ) -> "GetUsersMessage":
        """Creates a GetUsersMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetUsersMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetUsersMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetUsersMessage_PB


@serializable()
@final
class GetCandidatesMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)

    def _object2proto(self) -> GetCandidatesMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetUsersMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetCandidatesMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetCandidatesMessage_PB,
    ) -> "GetCandidatesMessage":
        """Creates a GetUsersMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetUsersMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetCandidatesMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetCandidatesMessage_PB


@serializable()
@final
class GetCandidatesResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        content: List[Dict],
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.content = content

    def _object2proto(self) -> GetCandidatesResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        msg = GetCandidatesResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
        )
        _ = [msg.content.append(serialize(content)) for content in self.content]
        return msg

    @staticmethod
    def _proto2object(
        proto: GetCandidatesResponse_PB,
    ) -> "GetCandidatesResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """
        return GetCandidatesResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=[_deserialize(content) for content in proto.content],
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetCandidatesResponse_PB


@serializable()
@final
class GetUsersResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        content: List[Dict],
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.content = content

    def _object2proto(self) -> GetUsersResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        msg = GetUsersResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
        )
        _ = [msg.content.append(serialize(content)) for content in self.content]
        return msg

    @staticmethod
    def _proto2object(
        proto: GetUsersResponse_PB,
    ) -> "GetUsersResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """
        return GetUsersResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=[_deserialize(content) for content in proto.content],
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetUsersResponse_PB


@serializable()
@final
class UpdateUserMessage(ImmediateSyftMessageWithReply):
    def __init__(  # nosec
        self,
        address: Address,
        reply_to: Address,
        user_id: Optional[int] = 0,
        msg_id: Optional[UID] = None,
        email: Optional[str] = "",
        password: Optional[str] = "",
        role: Optional[str] = "",
        groups: Optional[str] = "",
        budget: Optional[float] = None,
        name: Optional[str] = "",
        institution: Optional[str] = "",
        website: Optional[str] = "",
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.user_id = user_id
        self.email = email
        self.password = password
        self.role = role
        self.groups = groups
        self.name = name
        self.budget = budget
        self.institution = institution
        self.website = website

    def _object2proto(self) -> UpdateUserMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: UpdateUserMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return UpdateUserMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            user_id=self.user_id,
            email=self.email,
            password=self.password,
            role=self.role,
            groups=self.groups,
            budget=self.budget,
            institution=self.institution,
            website=self.website,
            name=self.name,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: UpdateUserMessage_PB,
    ) -> "UpdateUserMessage":
        """Creates a UpdateUserMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: UpdateUserMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return UpdateUserMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            user_id=proto.user_id,
            email=proto.email,
            password=proto.password,
            role=proto.role,
            budget=proto.budget,
            groups=proto.groups,
            name=proto.name,
            institution=proto.institution,
            website=proto.website,
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return UpdateUserMessage_PB


@serializable()
@final
class DeleteUserMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        user_id: int,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.user_id = user_id

    def _object2proto(self) -> DeleteUserMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: DeleteUserMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return DeleteUserMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            user_id=self.user_id,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: DeleteUserMessage_PB,
    ) -> "DeleteUserMessage":
        """Creates a DeleteUserMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: DeleteUserMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return DeleteUserMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            user_id=proto.user_id,
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return DeleteUserMessage_PB


@serializable()
@final
class SearchUsersMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
        email: Optional[str] = "",
        role: Optional[str] = "",
        groups: Optional[str] = "",
        name: Optional[str] = "",
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.email = email
        self.role = role
        self.groups = groups
        self.name = name

    def _object2proto(self) -> SearchUsersMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SearchUsersMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return SearchUsersMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            email=self.email,
            role=self.role,
            groups=self.groups,
            name=self.name,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: SearchUsersMessage_PB,
    ) -> "SearchUsersMessage":
        """Creates a SearchUsersMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SearchUsersMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return SearchUsersMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            email=proto.email,
            role=proto.role,
            groups=proto.groups,
            name=proto.name,
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return SearchUsersMessage_PB


@serializable()
@final
class SearchUsersResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        content: List[Dict],
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.content = content

    def _object2proto(self) -> SearchUsersResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return SearchUsersResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: SearchUsersResponse_PB,
    ) -> "SearchUsersResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return SearchUsersResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return SearchUsersResponse_PB



# stdlib
from collections import UserDict
from collections import UserList
from collections import UserString
from typing import NewType
from typing import Union

# relative
from .primitive_interface import PyPrimitive

NotImplementedType = NewType("NotImplementedType", type(NotImplemented))  # type: ignore
ImplementedInterfaces = Union[UserDict, UserList, UserString]
ImplementedSingletons = Union[type(None)]  # type: ignore
PrimitiveRetType = Union[
    PyPrimitive,
    NotImplementedType,
    ImplementedInterfaces,
    ImplementedSingletons,  # type: ignore
]
SyPrimitiveRet = NewType("SyPrimitiveRet", PrimitiveRetType)  # type: ignore



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/tensor/fixed_precision_tensor.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.tensor import (
    share_tensor_pb2 as proto_dot_core_dot_tensor_dot_share__tensor__pb2,
)
from syft.proto.core.tensor import (
    tensor_pb2 as proto_dot_core_dot_tensor_dot_tensor__pb2,
)
from syft.proto.lib.numpy import array_pb2 as proto_dot_lib_dot_numpy_dot_array__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n.proto/core/tensor/fixed_precision_tensor.proto\x12\x10syft.core.tensor\x1a\x1eproto/core/tensor/tensor.proto\x1a$proto/core/tensor/share_tensor.proto\x1a\x1bproto/lib/numpy/array.proto"\xc8\x01\n\x14\x46ixedPrecisionTensor\x12*\n\x06tensor\x18\x01 \x01(\x0b\x32\x18.syft.core.tensor.TensorH\x00\x12.\n\x05share\x18\x02 \x01(\x0b\x32\x1d.syft.core.tensor.ShareTensorH\x00\x12+\n\x05\x61rray\x18\x03 \x01(\x0b\x32\x1a.syft.lib.numpy.NumpyProtoH\x00\x12\x0c\n\x04\x62\x61se\x18\x04 \x01(\r\x12\x11\n\tprecision\x18\x05 \x01(\rB\x06\n\x04\x64\x61tab\x06proto3'
)


_FIXEDPRECISIONTENSOR = DESCRIPTOR.message_types_by_name["FixedPrecisionTensor"]
FixedPrecisionTensor = _reflection.GeneratedProtocolMessageType(
    "FixedPrecisionTensor",
    (_message.Message,),
    {
        "DESCRIPTOR": _FIXEDPRECISIONTENSOR,
        "__module__": "proto.core.tensor.fixed_precision_tensor_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.tensor.FixedPrecisionTensor)
    },
)
_sym_db.RegisterMessage(FixedPrecisionTensor)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _FIXEDPRECISIONTENSOR._serialized_start = 168
    _FIXEDPRECISIONTENSOR._serialized_end = 368
# @@protoc_insertion_point(module_scope)



"""SPDZ Protocol.

SPDZ mechanism used for multiplication Contains functions that are run at:

* the party that orchestrates the computation
* the parties that hold the shares
"""

# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import TYPE_CHECKING

# relative
from .....ast.klass import get_run_class_method
from ....tensor.smpc import utils
from ...store import CryptoPrimitiveProvider

EXPECTED_OPS = {"mul", "matmul"}

if TYPE_CHECKING:
    # relative
    from ....tensor.smpc.mpc_tensor import MPCTensor


def mul_master(
    x: MPCTensor, y: MPCTensor, op_str: str, **kwargs: Dict[Any, Any]
) -> MPCTensor:

    """Function that is executed by the orchestrator to multiply two secret values.

    Args:
        x (MPCTensor): First value to multiply with.
        y (MPCTensor): Second value to multiply with.
        op_str (str): Operation string.

    Raises:
        ValueError: If op_str not in EXPECTED_OPS.

    Returns:
        MPCTensor: Result of the multiplication.
    """

    # relative
    from ....tensor.tensor import TensorPointer

    parties = x.parties
    parties_info = x.parties_info

    shape_x = tuple(x.shape)  # type: ignore
    shape_y = tuple(y.shape)  # type: ignore

    ring_size = utils.get_ring_size(x.ring_size, y.ring_size)

    if ring_size != 2:
        # For ring_size 2 we generate those before hand
        CryptoPrimitiveProvider.generate_primitives(
            f"beaver_{op_str}",
            parties=parties,
            g_kwargs={
                "a_shape": shape_x,
                "b_shape": shape_y,
                "parties_info": parties_info,
            },
            p_kwargs={"a_shape": shape_x, "b_shape": shape_y},
            ring_size=ring_size,
        )

    # TODO: Should modify to parallel execution.
    if not isinstance(x.child[0], TensorPointer):
        res_shares = []
        for a, b in zip(x.child, y.child):
            if hasattr(a, "__mul__"):
                res_shares.append(a.__mul__(a, b, shape_x, shape_y, **kwargs))
    else:
        res_shares = []
        attr_path_and_name = f"{x.child[0].path_and_name}.__{op_str}__"
        op = get_run_class_method(attr_path_and_name, SMPC=True)
        for a, b in zip(x.child, y.child):
            res_shares.append(op(a, a, b, shape_x, shape_y, **kwargs))

    return res_shares  # type: ignore


def lt_master(x: MPCTensor, y: MPCTensor, op_str: str) -> MPCTensor:
    """Function that is executed by the orchestrator to multiply two secret values.

    Args:
        x (MPCTensor): First value to multiply with.
        y (MPCTensor): Second value to multiply with.
        op_str (str): Operation string.

    Raises:
        ValueError: If op_str not in EXPECTED_OPS.

    Returns:
        MPCTensor: Result of the multiplication.
    """
    # relative
    from ....tensor.smpc.mpc_tensor import MPCTensor
    from ....tensor.tensor import TensorPointer

    # diff = a - b
    # bit decomposition
    # sum carry adder
    # res = sign(diff)

    res_shares = x - y
    res_shares.block
    # time.sleep(2)
    msb = MSB(res_shares)
    tensor_shares = []
    final_shares = []

    if isinstance(x, MPCTensor):
        if isinstance(x.child[0], TensorPointer):
            for t1, t2 in zip(x.child, y.child):
                tensor_shares.append(t1.__lt__(t2))

            for p1, p2 in zip(tensor_shares, msb.child):
                p2.block
                final_shares.append(p1.mpc_swap(p2))

            msb.child = final_shares
    else:
        if isinstance(y.child[0], TensorPointer):  # type: ignore
            for t1 in y.child:
                tensor_shares.append(t1.__lt__(x))

            for p1, p2 in zip(tensor_shares, msb.child):
                p2.block
                final_shares.append(p1.mpc_swap(p2))

            msb.child = final_shares

    return msb


def MSB(x: MPCTensor) -> MPCTensor:
    # relative
    from ..aby3 import ABY3

    """Computes the MSB of the underlying share

    Args:
        x (MPCTensor): Input MPCTensor to compute MSB on.

    Returns:
        msb (MPCTensor): returns arithmetic shares of the MSB.
    """
    ring_size = 2**32  # TODO : Should extract ring_size elsewhere for generality.
    decomposed_shares = ABY3.bit_decomposition(x)

    for share in decomposed_shares:
        share.block

    msb_share = decomposed_shares[-1]
    msb = ABY3.bit_injection(msb_share, ring_size)

    return msb



# stdlib
import uuid

# relative
from ..tensor import AutogradTensor
from .op import Op


class CumSumOp(Op):
    def forward(self, x: AutogradTensor, y: AutogradTensor) -> None:  # type: ignore
        self.x = x
        self.y = y

    def _backward(self, grad: AutogradTensor, backprop_id: uuid.UUID) -> None:
        if self.x.requires_grad:
            pass

        if self.y.requires_grad:
            pass



"""This module contains `Class` attribute,an AST node representing a class."""

# stdlib
from enum import Enum
from enum import EnumMeta
import inspect
import sys
from types import ModuleType
from typing import Any
from typing import Callable as CallableT
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
import warnings

# relative
from .. import ast
from .. import lib
from ..core.common.group import VERIFYALL
from ..core.common.uid import UID
from ..core.node.common.action.action_sequence import ActionSequence
from ..core.node.common.action.get_or_set_property_action import GetOrSetPropertyAction
from ..core.node.common.action.get_or_set_property_action import PropertyActions
from ..core.node.common.action.run_class_method_action import RunClassMethodAction
from ..core.node.common.action.run_class_method_smpc_action import (
    RunClassMethodSMPCAction,
)
from ..core.node.common.action.save_object_action import SaveObjectAction
from ..core.node.common.node_service.resolve_pointer_type.resolve_pointer_type_messages import (
    ResolvePointerTypeMessage,
)
from ..core.node.common.util import upload_to_s3_using_presigned
from ..core.pointer.pointer import Pointer
from ..core.store.storeable_object import StorableObject
from ..logger import traceback_and_raise
from ..logger import warning
from ..util import aggressive_set_attr
from ..util import inherit_tags
from .callable import Callable


def _resolve_pointer_type(self: Pointer) -> Pointer:
    """Resolve pointer of the object.

    Creates a request on a pointer to validate and regenerate the current pointer type. This method
    is useful when dealing with AnyPointer or Union<types>Pointers, to retrieve the real pointer.

    The existing pointer will be deleted and a new one will be generated. The remote data won't
    be touched.

    Args:
        self: The pointer which will be validated.

    Returns:
        The new pointer, validated from the remote object.
    """
    # id_at_location has to be preserved
    id_at_location = getattr(self, "id_at_location", None)

    if id_at_location is None:
        traceback_and_raise(
            ValueError("Can't resolve a pointer that has no underlying object.")
        )

    cmd = ResolvePointerTypeMessage(
        id_at_location=id_at_location,
        address=self.client.address,
        reply_to=self.client.address,
    )

    # the path to the underlying type. It has to live in the AST
    real_type_path = self.client.send_immediate_msg_with_reply(msg=cmd).type_path
    new_pointer = self.client.lib_ast.query(real_type_path).pointer_type(
        client=self.client, id_at_location=id_at_location
    )

    # we disable the garbage collection message and then we delete the existing message.
    self.gc_enabled = False
    del self

    return new_pointer


def get_run_class_method(attr_path_and_name: str, SMPC: bool = False) -> CallableT:
    """Create a function for class method in `attr_path_and_name` for remote execution.

    Args:
        attr_path_and_name: The path of the class method.

    Returns:
        Function for the class method.

    Note:
        It might seem hugely un-necessary to have these methods nested in this way.
        However, it has to do with ensuring that the scope of `attr_path_and_name` is local
        and not global.

        If we do not put a `get_run_class_method` around `run_class_method` then
        each `run_class_method` will end up referencing the same `attr_path_and_name` variable
        and all methods will actually end up calling the same method.

        If, instead, we return the function object itself then it includes
        the current `attr_path_and_name` as an internal variable and when we call `get_run_class_method`
        multiple times it returns genuinely different methods each time with a different
        internal `attr_path_and_name` variable.
    """
    # relative
    from ..core.node.common.action import smpc_action_functions
    from ..core.node.common.action.smpc_action_functions import MAP_FUNC_TO_ACTION

    def run_class_smpc_method(
        __self: Any,
        *args: Tuple[Any, ...],
        **kwargs: Any,
    ) -> object:
        """Run remote class method on a SharePointer and get pointer to returned object.

        Args:
            *args: Args list of class method.
            **kwargs: Keyword args of class method.

        Returns:
            Pointer to object returned by class method.
        """
        seed_id_locations = kwargs.get("seed_id_locations", None)
        if seed_id_locations is None:
            raise ValueError(
                "There should be a `seed_id_locations` kwargs when doing an operation for MPCTensor"
            )

        kwargs["seed_id_locations"] = str(seed_id_locations)

        op = attr_path_and_name.split(".")[-1]
        id_at_location = smpc_action_functions.get_id_at_location_from_op(
            seed_id_locations, op
        )

        # we want to get the return type which matches the attr_path_and_name
        # so we ask lib_ast for the return type name that matches out
        # attr_path_and_name and then use that to get the actual pointer klass
        # then set the result to that pointer klass
        return_type_name = __self.client.lib_ast.query(
            attr_path_and_name
        ).return_type_name
        resolved_pointer_type = __self.client.lib_ast.query(return_type_name)
        result = resolved_pointer_type.pointer_type(client=__self.client)
        result.id_at_location = id_at_location

        # first downcast anything primitive which is not already PyPrimitive
        (
            downcast_args,
            downcast_kwargs,
        ) = lib.python.util.downcast_args_and_kwargs(args=args, kwargs=kwargs)

        # then we convert anything which isnt a pointer into a pointer
        pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
            args=downcast_args,
            kwargs=downcast_kwargs,
            client=__self.client,
            gc_enabled=False,
        )

        cmd = RunClassMethodSMPCAction(
            path=attr_path_and_name,
            _self=__self,
            args=pointer_args,
            kwargs=pointer_kwargs,
            id_at_location=result.id_at_location,
            address=__self.client.address,
        )
        __self.client.send_immediate_msg_without_reply(msg=cmd)

        inherit_tags(
            attr_path_and_name=attr_path_and_name,
            result=result,
            self_obj=__self,
            args=args,
            kwargs=kwargs,
        )

        return result

    def run_class_method(
        __self: Any,
        *args: Tuple[Any, ...],
        **kwargs: Any,
    ) -> object:
        """Run remote class method and get pointer to returned object.

        Args:
            *args: Args list of class method.
            **kwargs: Keyword args of class method.

        Returns:
            Pointer to object returned by class method.
        """

        # we want to get the return type which matches the attr_path_and_name
        # so we ask lib_ast for the return type name that matches out
        # attr_path_and_name and then use that to get the actual pointer klass
        # then set the result to that pointer klass
        return_type_name = __self.client.lib_ast.query(
            attr_path_and_name
        ).return_type_name
        resolved_pointer_type = __self.client.lib_ast.query(return_type_name)
        result = resolved_pointer_type.pointer_type(client=__self.client)

        # QUESTION can the id_at_location be None?
        result_id_at_location = getattr(result, "id_at_location", None)
        if result_id_at_location is not None:
            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(args=args, kwargs=kwargs)

            # then we convert anything which isnt a pointer into a pointer
            pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
                args=downcast_args,
                kwargs=downcast_kwargs,
                client=__self.client,
                gc_enabled=False,
            )

            cmd = RunClassMethodAction(
                path=attr_path_and_name,
                _self=__self,
                args=pointer_args,
                kwargs=pointer_kwargs,
                id_at_location=result_id_at_location,
                address=__self.client.address,
            )
            __self.client.send_immediate_msg_without_reply(msg=cmd)

        inherit_tags(
            attr_path_and_name=attr_path_and_name,
            result=result,
            self_obj=__self,
            args=args,
            kwargs=kwargs,
        )

        return result

    method_name = attr_path_and_name.rsplit(".", 1)[-1]
    if SMPC or (
        "ShareTensor" in attr_path_and_name and method_name in MAP_FUNC_TO_ACTION
    ):
        return run_class_smpc_method

    return run_class_method


def generate_class_property_function(
    attr_path_and_name: str, action: PropertyActions, map_to_dyn: bool
) -> CallableT:
    """Returns a function that handles action on property.

    Args:
        attr_path_and_name: The path of the property in AST.
        action: action to perform on property (GET | SET | DEL).

    Returns:
        Function to handle action on property.
    """

    def class_property_function(__self: Any, *args: Any, **kwargs: Any) -> object:
        """Handles remote action on property and returns pointer.

        Args:
            *args: Argument list.
            **kwargs: Keyword arguments.

        Returns:
            Pointer to the object returned.
        """
        # we want to get the return type which matches the attr_path_and_name
        # so we ask lib_ast for the return type name that matches out
        # attr_path_and_name and then use that to get the actual pointer klass
        # then set the result to that pointer klass
        return_type_name = __self.client.lib_ast.query(
            attr_path_and_name
        ).return_type_name
        resolved_pointer_type = __self.client.lib_ast.query(return_type_name)
        result = resolved_pointer_type.pointer_type(client=__self.client)
        # QUESTION can the id_at_location be None?
        result_id_at_location = getattr(result, "id_at_location", None)
        if result_id_at_location is not None:
            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(args=args, kwargs=kwargs)

            # then we convert anything which isnt a pointer into a pointer
            pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
                args=downcast_args, kwargs=downcast_kwargs, client=__self.client
            )

            cmd = GetOrSetPropertyAction(
                path=attr_path_and_name,
                id_at_location=result_id_at_location,
                address=__self.client.address,
                _self=__self,
                args=pointer_args,
                kwargs=pointer_kwargs,
                action=action,
                map_to_dyn=map_to_dyn,
            )
            __self.client.send_immediate_msg_without_reply(msg=cmd)

        if action == PropertyActions.GET:
            inherit_tags(
                attr_path_and_name=attr_path_and_name,
                result=result,
                self_obj=__self,
                args=args,
                kwargs=kwargs,
            )
        return result

    return class_property_function


def _get_request_config(self: Any) -> Dict[str, Any]:
    """Get config for request.

    Args:
        self: object.

    Returns:
        Config for request.
    """
    return {
        "request_block": True,
        "timeout_secs": 25,
        "delete_obj": False,
    }


def _set_request_config(self: Any, request_config: Dict[str, Any]) -> None:
    """Set config for request.

    Args:
        self: object.
        request_config: new config.
    """
    self.get_request_config = lambda: request_config


def wrap_iterator(attrs: Dict[str, Union[str, CallableT, property]]) -> None:
    """Add syft Iterator to `attrs['__iter__']`.

    Args:
        attrs: Dict of `Attribute`s of node.

    Raises:
        AttributeError: Base `__iter__` is not callable.
    """

    def wrap_iter(iter_func: CallableT) -> CallableT:
        """Create syft iterator for `iter_func`.

        Args:
            iter_func: Base Iterator.

        Returns:
            Wrapped Iterator.
        """

        def __iter__(self: Any) -> Iterable:
            """Create Syft Iterator for `iter_func`.

            Args:
                self: object to add iterator to.

            Raises:
                ValueError: Falied ot access __len__.

            Returns:
                Iterable: syft Iterator.
            """
            # relative
            from ..lib.python.iterator import Iterator

            if not hasattr(self, "__len__"):
                traceback_and_raise(
                    ValueError(
                        "Can't build a remote iterator on an object with no __len__."
                    )
                )

            try:
                data_len = self.__len__()
            except Exception:
                traceback_and_raise(
                    ValueError("Request to access data length rejected.")
                )

            return Iterator(_ref=iter_func(self), max_len=data_len)

        return __iter__

    attr_name = "__iter__"
    iter_target = attrs[attr_name]

    # skip if __iter__ has already been wrapped
    qual_name = getattr(iter_target, "__qualname__", None)
    if qual_name and "wrap_iter" in qual_name:
        return

    if not callable(iter_target):
        traceback_and_raise(AttributeError("Can't wrap a non callable iter attribute"))
    else:
        iter_func: CallableT = iter_target
    attrs[attr_name] = wrap_iter(iter_func)


def wrap_len(attrs: Dict[str, Union[str, CallableT, property]]) -> None:
    """Add method to access pointer len to `attr[__len__]`.

    Args:
        attrs: Dict of `Attribute`s of node.

    Raises:
        AttributeError: Base `__len__` is not callable.
    """

    def wrap_len(len_func: CallableT) -> CallableT:
        """Add wrapper function for `len_func`.

        Args:
            len_func: Base len function.

        Returns:
            Wrapped len function.
        """

        def __len__(self: Any) -> int:
            """Access len of pointer obj.

            Args:
                self: object to add iterator to.

            Returns:
                int: length of object.

            Raises:
                ValueError: Request to access data length rejected.
            """
            data_len_ptr = len_func(self)
            try:
                print(self.get_request_config())
                data_len = data_len_ptr.get(**self.get_request_config())

                if data_len is None:
                    raise Exception

                return data_len
            except Exception:
                traceback_and_raise(
                    ValueError("Request to access data length rejected.")
                )

        return __len__

    attr_name = "__len__"
    len_target = attrs[attr_name]

    if not callable(len_target):
        traceback_and_raise(
            AttributeError("Can't wrap a non callable __len__ attribute")
        )
    else:
        len_func: CallableT = len_target

    attrs["len"] = len_func
    attrs[attr_name] = wrap_len(len_func)


def attach_tags(obj: object, tags: List[str]) -> None:
    """Add tags to the object.

    Args:
        obj: Object to add tags to.
        tags: List of tags.

    Raises:
        AttributeError: Cannot add tags to object.
    """
    try:
        obj.tags = sorted(set(tags), key=tags.index)  # type: ignore
    except AttributeError:
        warning(f"can't attach new attribute `tags` to {type(obj)} object.")


def attach_description(obj: object, description: str) -> None:
    """Add description to the object.

    Args:
        obj: Object to add description to.
        description: Description.

    Raises:
        AttributeError: Cannot add description to object.
    """
    try:
        obj.description = description  # type: ignore
    except AttributeError:
        warning(f"can't attach new attribute `description` to {type(obj)} object.")


class Class(Callable):
    """A Class attribute represents a class."""

    def __init__(
        self,
        path_and_name: str,
        parent: ast.attribute.Attribute,
        object_ref: Union[Callable, CallableT],
        return_type_name: Optional[str],
        client: Optional[Any],
    ) -> None:
        """Base constructor for Class Attribute.

        Args:
            path_and_name: The path for the current node, e.g. `syft.lib.python.List`.
            parent: The parent node is needed when solving `EnumAttributes`.
            object_ref: The actual python object for which the computation is being made.
            return_type_name: The return type name of given action as a string with its full path.
            client: The client for which all computation is being executed.
        """
        super().__init__(
            path_and_name=path_and_name,
            object_ref=object_ref,
            return_type_name=return_type_name,
            client=client,
            parent=parent,
        )
        if self.path_and_name is not None:
            self.pointer_name = self.path_and_name.split(".")[-1] + "Pointer"

    @property
    def pointer_type(self) -> Union[Callable, CallableT]:
        """Get pointer type of Class Attribute.

        Returns:
            `pointer_type` of the object.
        """
        return getattr(self, self.pointer_name)

    def create_pointer_class(self) -> None:
        """Create pointer type for object."""
        attrs: Dict[str, Union[str, CallableT, property]] = {}
        for attr_name, attr in self.attrs.items():
            attr_path_and_name: Optional[str] = getattr(attr, "path_and_name", None)

            if attr_path_and_name is None:
                raise Exception(f"Missing path_and_name in {self.attrs}")

            # attr_path_and_name None
            if isinstance(attr, ast.callable.Callable):
                attrs[attr_name] = get_run_class_method(attr_path_and_name)
            elif isinstance(attr, ast.property.Property):
                prop = property(
                    generate_class_property_function(
                        attr_path_and_name, PropertyActions.GET, map_to_dyn=False
                    )
                )

                prop = prop.setter(
                    generate_class_property_function(
                        attr_path_and_name, PropertyActions.SET, map_to_dyn=False
                    )
                )
                prop = prop.deleter(
                    generate_class_property_function(
                        attr_path_and_name, PropertyActions.DEL, map_to_dyn=False
                    )
                )
                attrs[attr_name] = prop
            elif isinstance(attr, ast.dynamic_object.DynamicObject):
                prop = property(
                    generate_class_property_function(
                        attr_path_and_name, PropertyActions.GET, map_to_dyn=True
                    )
                )

                prop = prop.setter(
                    generate_class_property_function(
                        attr_path_and_name, PropertyActions.SET, map_to_dyn=True
                    )
                )
                prop = prop.deleter(
                    generate_class_property_function(
                        attr_path_and_name, PropertyActions.DEL, map_to_dyn=True
                    )
                )
                attrs[attr_name] = prop
            if attr_name == "__len__":
                wrap_len(attrs)

            if getattr(attr, "return_type_name", None) == "syft.lib.python.Iterator":
                wrap_iterator(attrs)

        attrs["get_request_config"] = _get_request_config
        attrs["set_request_config"] = _set_request_config
        attrs["resolve_pointer_type"] = _resolve_pointer_type

        fqn = "Pointer"

        if self.path_and_name is not None:
            fqn = self.path_and_name + fqn

        new_class_name = f"syft.proxy.{fqn}"
        parts = new_class_name.split(".")
        name = parts.pop(-1)
        attrs["__name__"] = name
        attrs["__module__"] = ".".join(parts)

        # if the object already has a pointer class specified, use that instead of creating
        # an empty subclass of Pointer
        if hasattr(self.object_ref, "PointerClassOverride"):

            klass_pointer = getattr(self.object_ref, "PointerClassOverride")
            for key, val in attrs.items():

                # only override functioanlity of AST attributes if they
                # don't already exist on the PointerClassOverride class
                # (the opposite of inheritance)
                if not hasattr(klass_pointer, key):
                    setattr(klass_pointer, key, val)
                else:
                    # TODO: cache attribute in backup_ location so that we can use them if we want
                    pass

        # no specific pointer class found, let's make an empty subclass of Pointer instead
        else:
            klass_pointer = type(self.pointer_name, (Pointer,), attrs)

        setattr(klass_pointer, "path_and_name", self.path_and_name)
        setattr(self, self.pointer_name, klass_pointer)

        module_type = type(sys)

        # syft absolute
        import syft

        parent = syft
        for part in parts[1:]:
            if part not in parent.__dict__:
                parent.__dict__[part] = module_type(name=part)
            parent = parent.__dict__[part]
        parent.__dict__[name] = klass_pointer

    def store_init_args(outer_self: Any) -> None:
        """
        Stores args and kwargs of outer_self init by wrapping the init method.
        """

        def init_wrapper(self: Any, *args: List[Any], **kwargs: Dict[Any, Any]) -> None:
            outer_self.object_ref._wrapped_init(self, *args, **kwargs)
            self._init_args = args
            self._init_kwargs = kwargs

        # If _wrapped_init already exists, create_init_method is already called once
        # and does not need to wrap __init__ again.
        if not hasattr(outer_self.object_ref, "_wrapped_init"):
            outer_self.object_ref._wrapped_init = outer_self.object_ref.__init__
            outer_self.object_ref.__init__ = init_wrapper

    def create_send_method(outer_self: Any) -> None:
        """Add `send` method to `outer_self.object_ref`."""

        def send(
            self: Any,
            client: Any,
            pointable: bool = True,
            description: str = "",
            tags: Optional[List[str]] = None,
            searchable: Optional[bool] = None,
            id_at_location_override: Optional[UID] = None,
            chunk_size: Optional[int] = None,
            send_to_blob_storage: bool = False,
            **kwargs: Dict[str, Any],
        ) -> Union[Pointer, Tuple[Pointer, SaveObjectAction]]:

            """Send obj to client and return pointer to the object.

            Args:
                self: Object to be sent.
                client: Client to send object to.
                pointable:
                description: Description for the object to send.
                tags: Tags for the object to send.

            Returns:
                Pointer to sent object.

            Note:
                `searchable` is deprecated please use `pointable` in the future.
            """
            if searchable is not None:
                msg = "`searchable` is deprecated please use `pointable` in future"
                warning(msg, print=True)
                warnings.warn(
                    msg,
                    DeprecationWarning,
                )
                pointable = searchable

            chunk_size = chunk_size if chunk_size is not None else 536870912  # 500 MB

            if not hasattr(self, "id"):
                try:
                    self.id = UID()
                except AttributeError:
                    pass

            # if `tags` is passed in, use it; else, use obj_tags
            obj_tags = getattr(self, "tags", [])
            tags = tags if tags else []
            tags = tags if tags else obj_tags

            # if `description` is passed in, use it; else, use obj_description
            obj_description = getattr(self, "description", "")
            description = description if description else obj_description

            # TODO: Allow Classes to opt out in the AST like Pandas where the properties
            # would break their dict attr usage
            # Issue: https://github.com/OpenMined/PySyft/issues/5322
            if outer_self.pointer_name not in {"DataFramePointer", "SeriesPointer"}:
                attach_tags(self, tags)
                attach_description(self, description)

            if id_at_location_override is not None:
                id_at_location = id_at_location_override
            else:
                id_at_location = UID()

            if hasattr(self, "init_pointer"):
                constructor = self.init_pointer
            else:
                constructor = getattr(outer_self, outer_self.pointer_name)

            # Step 1: create pointer which will point to result
            ptr = constructor(
                client=client,
                id_at_location=id_at_location,
                tags=tags,
                description=description,
            )

            ptr._pointable = pointable

            if pointable:
                ptr.gc_enabled = False
            else:
                ptr.gc_enabled = True

            if send_to_blob_storage:
                store_data = upload_to_s3_using_presigned(
                    client=client,
                    data=self,
                    chunk_size=chunk_size,
                    asset_name=id_at_location.no_dash,
                )
            else:
                store_data = self

            # Step 6: create message which contains object to send
            storable = StorableObject(
                id=ptr.id_at_location,
                data=store_data,
                tags=tags,
                description=description,
                search_permissions={VERIFYALL: None} if pointable else {},
            )
            obj_msg = SaveObjectAction(obj=storable, address=client.address)

            immediate = kwargs.get("immediate", True)

            if immediate:
                # Step 7: send message
                client.send_immediate_msg_without_reply(msg=obj_msg)

                # Step 8: return pointer
                return ptr
            else:
                return ptr, obj_msg

        aggressive_set_attr(obj=outer_self.object_ref, name="send", attr=send)

    def create_storable_object_attr_convenience_methods(outer_self: Any) -> None:
        """Add methods to set tag and description to `outer_self.object_ref`."""

        def tag(self: Any, *tags: Tuple[Any, ...]) -> object:
            """Add tags to object.

            Args:
                self: object to add tags to.
                *tags: List of tags to add.

            Returns:
                object.
            """
            attach_tags(self, tags)  # type: ignore
            return self

        def describe(self: Any, description: str) -> object:
            """Add description to object.

            Args:
                self: object to add description to.
                description: Description to add.

            Returns:
                object.
            """
            attach_description(self, description)
            return self

        aggressive_set_attr(obj=outer_self.object_ref, name="tag", attr=tag)
        aggressive_set_attr(obj=outer_self.object_ref, name="describe", attr=describe)

    def add_path(
        self,
        path: Union[str, List[str]],
        index: int,
        return_type_name: Optional[str] = None,
        framework_reference: Optional[ModuleType] = None,
        is_static: bool = False,
    ) -> None:
        """The add_path method adds new nodes in AST based on type of current node and type of object to be added.

        Args:
            path: The node path added in AST, e.g. `syft.lib.python.List` or ["syft", "lib", "python", "List].
            index: The associated position in the path for the current node.
            framework_reference: The Python framework in which we can resolve same path to obtain Python object.
            return_type_name: The return type name of the given action as a string with its full path.
            is_static: If the queried object is static, it has to be found on AST itself, not on an existing pointer.
        """
        if index >= len(path) or path[index] in self.attrs:
            return

        _path: List[str] = path.split(".") if isinstance(path, str) else path
        attr_ref = getattr(self.object_ref, _path[index])

        class_is_enum = isinstance(self.object_ref, EnumMeta)

        if (
            inspect.isfunction(attr_ref)
            or inspect.isbuiltin(attr_ref)
            or inspect.ismethod(attr_ref)
            or inspect.ismethoddescriptor(attr_ref)
        ):
            super().add_path(_path, index, return_type_name)
        if isinstance(attr_ref, Enum) and class_is_enum:
            enum_attribute = ast.enum.EnumAttribute(
                path_and_name=".".join(_path[: index + 1]),
                return_type_name=return_type_name,
                client=self.client,
                parent=self,
            )
            setattr(self, _path[index], enum_attribute)
            self.attrs[_path[index]] = enum_attribute

        elif inspect.isdatadescriptor(attr_ref) or inspect.isgetsetdescriptor(attr_ref):
            self.attrs[_path[index]] = ast.property.Property(
                path_and_name=".".join(_path[: index + 1]),
                object_ref=attr_ref,
                return_type_name=return_type_name,
                client=self.client,
                parent=self,
            )
        elif not callable(attr_ref):
            static_attribute = ast.static_attr.StaticAttribute(
                path_and_name=".".join(_path[: index + 1]),
                return_type_name=return_type_name,
                client=self.client,
                parent=self,
            )
            setattr(self, _path[index], static_attribute)
            self.attrs[_path[index]] = static_attribute

    def add_dynamic_object(self, path_and_name: str, return_type_name: str) -> None:
        self.attrs[
            path_and_name.rsplit(".", maxsplit=1)[-1]
        ] = ast.dynamic_object.DynamicObject(
            path_and_name=path_and_name,
            return_type_name=return_type_name,
            client=self.client,
            parent=self,
        )

    def __getattribute__(self, item: str) -> Any:
        """Get pointer to attribute.

        Args:
            item: Attribute.

        Returns:
            Pointer to the attribute.
        """
        # self.apply_node_changes()
        try:
            target_object = super().__getattribute__(item)

            if isinstance(target_object, ast.static_attr.StaticAttribute):
                return target_object.get_remote_value()

            if isinstance(target_object, ast.enum.EnumAttribute):
                target_object_ptr = target_object.get_remote_enum_attribute()
                target_object_ptr.is_enum = True
                return target_object_ptr

            return target_object
        except Exception as e:
            # TODO: this gets really chatty when doing SMPC mulitplication. Figure out why.
            # critical(
            #     f"{self.path_and_name}__getattribute__[{item}] failed. If you
            #     are trying to access an EnumAttribute or a "
            #     "StaticAttribute, be sure they have been added to the AST. Falling back on"
            #     "__getattr__ to search in self.attrs for the requested field."
            # )
            traceback_and_raise(e)

    def __getattr__(self, item: str) -> Any:
        """Get value of attribute `item` of the object.

        Args:
            item: Attribute.

        Raises:
            KeyError: If attribute `item` is not present.

        Returns:
            Value of the attribute.
        """
        attrs = super().__getattribute__("attrs")
        if item not in attrs:
            if item == "__name__":
                # return the pointer name if __name__ is missing
                return self.pointer_name
            traceback_and_raise(
                KeyError(
                    f"__getattr__ failed, {item} is not present on the "
                    f"object, nor the AST attributes!"
                )
            )
        return attrs[item]

    def __setattr__(self, key: str, value: Any) -> None:
        """Change value of attribute `key` to `value`.

        Args:
            key: name of attribute to change.
            value: value to change attribute `key` to.
        """
        # self.apply_node_changes()

        if hasattr(super(), "attrs"):
            attrs = super().__getattribute__("attrs")
            if key in attrs:
                target_object = self.attrs[key]
                if isinstance(target_object, ast.static_attr.StaticAttribute):
                    return target_object.set_remote_value(value)

        return super().__setattr__(key, value)


# TODO: this should move out of AST into a util somewhere? or osmething related to Pointer
def pointerize_args_and_kwargs(
    args: Union[List[Any], Tuple[Any, ...]],
    kwargs: Dict[Any, Any],
    client: Any,
    gc_enabled: bool = True,
) -> Tuple[List[Any], Dict[Any, Any]]:
    """Get pointers to args and kwargs.

    Args:
        args: List of arguments.
        kwargs: Dict of Keyword arguments.
        client: Client node.

    Returns:
        Tuple of args and kwargs with pointer to values.
    """
    # When we try to send params to a remote function they need to be pointers so
    # that they can be serialized and fetched from the remote store on arrival
    # this ensures that any args which are passed in from the user side are first
    # converted to pointers and sent then the pointer values are used for the
    # method invocation
    obj_lst = []
    pointer_args = []
    pointer_kwargs = {}
    for arg in args:
        # check if its already a pointer
        if not isinstance(arg, Pointer):
            arg_ptr, obj = arg.send(client, pointable=not gc_enabled, immediate=False)
            obj_lst.append(obj)
            pointer_args.append(arg_ptr)
        else:
            pointer_args.append(arg)
            arg.gc_enabled = gc_enabled

    for k, arg in kwargs.items():
        # check if its already a pointer
        if not isinstance(arg, Pointer):
            arg_ptr, obj = arg.send(client, pointable=not gc_enabled, immediate=False)
            obj_lst.append(obj)
            pointer_kwargs[k] = arg_ptr
        else:
            pointer_kwargs[k] = arg

    msg = ActionSequence(obj_lst=obj_lst, address=client.address)

    # send message to client
    client.send_immediate_msg_without_reply(msg=msg)

    return pointer_args, pointer_kwargs



# stdlib
from typing import Any

# third party
from sqlalchemy.ext.declarative import as_declarative


@as_declarative()
class Base:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    id: Any
    __name__: str



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
import numpy as np

# relative
from ...proto.core.tensor.fixed_precision_tensor_pb2 import (
    FixedPrecisionTensor as FixedPrecisionTensor_PB,
)  # type: ignore
from ..common.serde.deserialize import _deserialize as deserialize
from ..common.serde.serializable import serializable
from ..common.serde.serialize import _serialize as serialize
from .passthrough import PassthroughTensor  # type: ignore


@serializable()
class FixedPrecisionTensor(PassthroughTensor):
    def __init__(
        self, value: Optional[Any] = None, base: int = 10, precision: int = 3
    ) -> None:
        self._base = base
        self._precision = precision
        self._scale = base**precision
        if value is not None:
            fpt_value = self._scale * value
            encoded_value = fpt_value.astype(np.int64)
            super().__init__(encoded_value)

    def decode(self) -> Any:
        correction = (self.child < 0).astype(np.int64)
        dividend = self.child // self._scale - correction
        remainder = self.child % self._scale
        remainder += (remainder == 0).astype(np.int64) * self._scale * correction
        value = dividend.astype(np.float32) + remainder.astype(np.float32) / self._scale
        return value

    def __add__(self, other: Any) -> FixedPrecisionTensor:
        res = FixedPrecisionTensor(base=self._base, precision=self._precision)
        res.child = self.child + other.child
        return res

    def __sub__(self, other: Any) -> FixedPrecisionTensor:
        res = FixedPrecisionTensor(base=self._base, precision=self._precision)
        res.child = self.child - other.child
        return res

    def _object2proto(self) -> FixedPrecisionTensor_PB:
        # relative
        from .share_tensor import ShareTensor
        from .tensor import Tensor

        if isinstance(self.child, Tensor):
            return FixedPrecisionTensor_PB(
                tensor=serialize(self.child), base=self._base, precision=self._precision
            )
        elif isinstance(self.child, ShareTensor):
            return FixedPrecisionTensor_PB(
                share=serialize(self.child), base=self._base, precision=self._precision
            )

        return FixedPrecisionTensor_PB(
            array=serialize(self.child), base=self._base, precision=self._precision
        )

    @staticmethod
    def _proto2object(proto: FixedPrecisionTensor_PB) -> FixedPrecisionTensor:
        res = FixedPrecisionTensor(base=proto.base, precision=proto.precision)

        # Put it manually since we send it already encoded
        if proto.HasField("tensor"):
            res.child = deserialize(proto.tensor)
        elif proto.HasField("share"):
            res.child = deserialize(proto.share)
        else:
            res.child = deserialize(proto.array)
        return res

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return FixedPrecisionTensor_PB



# -*- coding: utf-8 -*-
"""This module contains utility funtions for Syft's AST submodule."""

# stdlib
from typing import List as TypeList

# third party
import torch

module_type = type(torch)
func_type = type(lambda x: x)
builtin_func_type = type(torch.ones)
class_type = type(func_type)


def unsplit(list_of_things: TypeList[str], separator: str = ".") -> str:
    """The unsplit method joins different parts of a node path using dot separator.

    Args:
        list_of_things: list of string to join.
        separator: Separator character/string. (default ".")

    Returns:
        str: string after join.
    """
    return separator.join(list_of_things)



# stdlib
from typing import Optional

# relative
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..node.abstract.node import AbstractNodeClient
from .scalar.abstract.scalar import Scalar  # noqa: 401
from .scalar.gamma_scalar import GammaScalar  # noqa: 401
from .scalar.intermediate_phi_scalar import IntermediatePhiScalar  # noqa: 401
from .scalar.phi_scalar import PhiScalar  # noqa: 401


def create_adp_ast(client: Optional[AbstractNodeClient] = None) -> Globals:
    ast = Globals(client)

    modules = ["syft", "syft.core", "syft.core.adp"]
    classes = [
        ("syft.core.adp.GammaScalar", "syft.core.adp.Scalar", GammaScalar),
        ("syft.core.adp.Scalar", "syft.core.adp.Scalar", Scalar),
        ("syft.core.adp.PhiScalar", "syft.core.adp.PhiScalar", PhiScalar),
        (
            "syft.core.adp.IntermediatePhiScalar",
            "syft.core.adp.IntermediatePhiScalar",
            IntermediatePhiScalar,
        ),
    ]

    methods = [
        # Scalar
        ("syft.core.adp.Scalar.max_val", "syft.lib.python.Float"),
        ("syft.core.adp.Scalar.min_val", "syft.lib.python.Float"),
        ("syft.core.adp.Scalar.value", "syft.lib.python.Float"),
        ("syft.core.adp.Scalar.publish", "syft.lib.python.Float"),
        # PhiScalar
        (
            "syft.core.adp.PhiScalar.__add__",
            "syft.core.adp.IntermediatePhiScalar",
        ),  # Union?
        (
            "syft.core.adp.PhiScalar.__mul__",
            "syft.core.adp.IntermediatePhiScalar",
        ),  # Union?
        (
            "syft.core.adp.PhiScalar.__radd__",
            "syft.core.adp.IntermediatePhiScalar",
        ),  # Union?
        (
            "syft.core.adp.PhiScalar.__rmul__",
            "syft.core.adp.IntermediatePhiScalar",
        ),  # Union?
        (
            "syft.core.adp.PhiScalar.__rsub__",
            "syft.core.adp.IntermediatePhiScalar",
        ),  # Union?
        (
            "syft.core.adp.PhiScalar.__sub__",
            "syft.core.adp.IntermediatePhiScalar",
        ),  # Union?
        ("syft.core.adp.PhiScalar.gamma", "syft.core.adp.GammaScalar"),
        (
            "syft.core.adp.PhiScalar.input_entities",
            "syft.lib.python.List",
        ),  # TypeList[Entity]
        (
            "syft.core.adp.PhiScalar.input_polys",
            "syft.lib.python.Set",
        ),  # TypeSet[BasicSymbol]
        (
            "syft.core.adp.PhiScalar.input_scalars",
            "syft.lib.python.List",
        ),  # TypeList[Union[PhiScalar, GammaScalar]]
        ("syft.core.adp.PhiScalar.max_lipschitz", "syft.lib.python.Float"),
        ("syft.core.adp.PhiScalar.max_lipschitz_wrt_entity", "syft.lib.python.Float"),
        ("syft.core.adp.PhiScalar.max_val", "syft.lib.python.Float"),
        ("syft.core.adp.PhiScalar.min_val", "syft.lib.python.Float"),
        ("syft.core.adp.PhiScalar.value", "syft.lib.python.Float"),
        ("syft.core.adp.PhiScalar.publish", "syft.lib.python.Float"),
        # IntermediatePhiScalar
        ("syft.core.adp.IntermediatePhiScalar.gamma", "syft.core.adp.GammaScalar"),
        (
            "syft.core.adp.IntermediatePhiScalar.__add__",
            "syft.core.adp.IntermediatePhiScalar",
        ),  # Union?
        (
            "syft.core.adp.IntermediatePhiScalar.__mul__",
            "syft.core.adp.IntermediatePhiScalar",
        ),  # Union?
        (
            "syft.core.adp.IntermediatePhiScalar.__radd__",
            "syft.core.adp.IntermediatePhiScalar",
        ),  # Union?
        (
            "syft.core.adp.IntermediatePhiScalar.__rmul__",
            "syft.core.adp.IntermediatePhiScalar",
        ),  # Union?
        (
            "syft.core.adp.IntermediatePhiScalar.__rsub__",
            "syft.core.adp.IntermediatePhiScalar",
        ),  # Union?
        (
            "syft.core.adp.IntermediatePhiScalar.__sub__",
            "syft.core.adp.IntermediatePhiScalar",
        ),  # Union?
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast



# future
from __future__ import annotations

# stdlib
from typing import List as TypeList
from typing import Optional
from typing import Set as TypeSet
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from pymbolic.interop.sympy import PymbolicToSympyMapper
from pymbolic.mapper.evaluator import EvaluationMapper as EM
from sympy.core.basic import Basic as BasicSymbol

# relative
from .....proto.core.adp.scalar_pb2 import IntermediateScalar as IntermediateScalar_PB
from ....common import UID
from ....common.serde.serializable import serializable
from ...entity import Entity
from ...search import GetSymbolsMapper
from ...search import flatten_and_maximize_poly
from ...search import ssid2obj
from .scalar import Scalar


@serializable()
class IntermediateScalar(Scalar):
    """
    Serializable Scalar class that supports polynomial representations of data.
    It is assumed that IntermediateScalar is immutable and therefore we will cache
    the sympoly creation because poly wont change
    """

    def __init__(self, poly: BasicSymbol, id: Optional[UID] = None) -> None:
        self.poly = poly
        self.id = id if id else UID()
        self._sympoly: Optional[BasicSymbol] = None

        # only initialize these if they aren't already set
        if not hasattr(self, "_min_val"):
            self._min_val: Optional[float] = None
        if not hasattr(self, "_max_val"):
            self._max_val: Optional[float] = None

    @property
    def sympoly(self) -> BasicSymbol:
        """Sympy version of self.poly"""
        if self._sympoly is None:
            self._sympoly = PymbolicToSympyMapper()(self.poly)
        return self._sympoly

    def __mul__(self, other: IntermediateScalar) -> IntermediateScalar:
        raise NotImplementedError

    def __rmul__(self, other: IntermediateScalar) -> IntermediateScalar:
        return self * other

    def __add__(self, other: IntermediateScalar) -> IntermediateScalar:
        raise NotImplementedError

    def __radd__(self, other: IntermediateScalar) -> IntermediateScalar:
        return self + other

    def __sub__(self, other: IntermediateScalar) -> IntermediateScalar:
        raise NotImplementedError

    def __rsub__(self, other: IntermediateScalar) -> IntermediateScalar:
        return other - self  # subtraction is not commutative

    @property
    def input_scalars(self) -> TypeList[Scalar]:
        """Return a list of the PhiScalar & GammaScalar objects used to create this Scalar."""
        phi_gamma_scalars: TypeList[Scalar] = list()
        for free_symbol in self.input_polys:
            ssid = str(free_symbol)
            phi_gamma_scalars.append(ssid2obj[ssid])
        return phi_gamma_scalars

    @property
    def input_entities(self) -> Union[TypeList[Entity], TypeList[None]]:
        """Return a list of the entities involved in the creation of this scalar object."""
        return list(set([x.entity for x in self.input_scalars]))  # type: ignore

    @property
    def input_polys(self) -> TypeSet[BasicSymbol]:
        """Use a mapper object to return the unique set of polynomials"""
        mapper = GetSymbolsMapper()  # type: ignore
        mapper(self.poly)
        return mapper.free_symbols

    @property
    def max_val(self) -> Optional[float]:
        # TODO: Verify that his doesnt change anything with budget spend
        if self._max_val is not None:
            return self._max_val

        if self.poly is not None:
            results = flatten_and_maximize_poly(-self.poly)
            if len(results) >= 1:
                self._max_val = float(-results[-1].fun)
                return self._max_val
        return None

    @property
    def min_val(self) -> Optional[float]:
        # TODO: Verify that his doesnt change anything with budget spend
        if self._min_val is not None:
            return self._min_val

        if self.poly is not None:
            results = flatten_and_maximize_poly(self.poly)
            if len(results) >= 1:
                self._min_val = float(results[-1].fun)
                return self._min_val
        return None

    @property
    def value(self) -> Optional[float]:

        if hasattr(self, "_value_cache"):
            return self._value_cache  # type: ignore

        if self.poly is not None:
            result = EM(
                context={obj.poly.name: obj.value for obj in self.input_scalars}  # type: ignore
            )(self.poly)
            return float(result)
        return None

    def _object2proto(self) -> IntermediateScalar_PB:
        return IntermediateScalar_PB(
            id=self.id._object2proto(),
            # poly=self._poly if self._poly is not None else None,
        )

    @staticmethod
    def _proto2object(proto: IntermediateScalar_PB) -> "IntermediateScalar":
        intermediate_scalar = IntermediateScalar(
            id=UID._proto2object(proto.id),
            poly=None
            # poly=sy.deserialize(proto.poly)
        )
        return intermediate_scalar

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return IntermediateScalar_PB



# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

# third party
from nacl.signing import VerifyKey
from pydantic import BaseModel
from pydantic.error_wrappers import ValidationError as PydanticValidationError

# relative
from .....common.message import ImmediateSyftMessage
from .....common.message import SignedMessage
from .....common.uid import UID
from .....io.address import Address
from ....abstract.node_service_interface import NodeServiceInterface
from ....common.exceptions import AuthorizationError
from ....common.exceptions import BadPayloadException
from ....common.exceptions import PermissionsNotDefined


# Inner Payload message using Pydantic.
class Payload(BaseModel):
    class Config:
        orm_mode = True


class RequestPayload(Payload):
    pass


class ReplyPayload(Payload):
    pass


class NewSyftMessage(ImmediateSyftMessage):
    """A base class from which all message classes should inherit.

    Note:
        This will eventually replace the old `SyftMessage` class.
    """

    __attr_allowlist__ = ["id", "address", "reply_to", "reply", "msg_id", "kwargs"]

    signed_type = SignedMessage
    request_payload_type = RequestPayload
    reply_payload_type = ReplyPayload

    def __init__(
        self,
        address: Address,
        kwargs: Optional[Dict[str, Any]] = None,
        msg_id: Optional[UID] = None,
        reply_to: Optional[Address] = None,
        reply: bool = False,
    ) -> None:
        super().__init__(address=address, msg_id=msg_id)
        self.reply_to = reply_to
        self.reply = reply
        self.kwargs = kwargs if kwargs else {}

    @property
    def payload(self) -> Payload:
        kwargs_dict = {}

        if hasattr(self.kwargs, "upcast"):
            kwargs_dict = self.kwargs.upcast()  # type: ignore
        else:
            kwargs_dict = self.kwargs  # type: ignore

        try:
            # If it's not a reply message then load kwargs as a proper request payload.
            if not self.reply:
                return self.request_payload_type(**kwargs_dict)
            # If it's a reply message, then load kwargs as a proper reply payload.
            else:
                return self.reply_payload_type(**kwargs_dict)
        except PydanticValidationError:
            raise BadPayloadException

    def run(
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:
        raise NotImplementedError

    def get_permissions(self) -> List:
        """Returns the list of permission classes applicable to the given message."""
        raise NotImplementedError

    def check_permissions(
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> None:
        """Check if the user has relevant permissions to run this message.

        Args:
            node (NodeServiceInterface): node interface used to invoke this message.
            verify_key (Optional[VerifyKey], optional): user signed verification key. Defaults to None.

        Raises:
            AuthorizationError: Error when one of the permission is denied.
        """

        if not len(self.get_permissions()):
            raise PermissionsNotDefined

        for permission_class in self.get_permissions():
            if not permission_class().has_permission(
                msg=self, node=node, verify_key=verify_key
            ):
                raise AuthorizationError(
                    f"You don't have access to perform {self} action."
                )



# stdlib
from typing import Dict as TypeDict
from typing import List as TypeList
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from ......lib.python import List  # type: ignore
from ......logger import traceback_and_raise  # type: ignore
from .....adp.publish import publish  # type: ignore
from .....common.uid import UID  # type: ignore
from .....store.storeable_object import StorableObject  # type: ignore
from .....tensor.tensor import PassthroughTensor  # type: ignore
from ....abstract.node import AbstractNode  # type: ignore
from ..node_service import ImmediateNodeServiceWithoutReply  # type: ignore
from .publish_messages import PublishScalarsAction  # type: ignore


class PublishScalarsService(ImmediateNodeServiceWithoutReply):
    @staticmethod
    def process(
        node: AbstractNode, msg: PublishScalarsAction, verify_key: VerifyKey
    ) -> None:

        print("PublishScalarsService:28")

        # get scalar objects from store
        results = List()
        for publish_id in msg.publish_ids_at_location:
            print("PublishScalarsService:33")
            print(publish_id)
            try:
                print(
                    "PublishScalarsService:36: TRY: publish_object = node.store.get(publish_id)"
                )
                publish_object = node.store.get(publish_id)
                print(
                    "PublishScalarsService:38: SUCCESS: publish_object = node.store.get(publish_id)"
                )
                if isinstance(publish_object.data, PassthroughTensor):
                    print(
                        "PublishScalarsService:40: TRY: publish_object.data.publish()"
                    )
                    result = publish_object.data.publish(
                        acc=node.acc, sigma=msg.sigma, user_key=verify_key
                    )
                    print(
                        "PublishScalarsService:44: SUCCESS: publish_object.data.publish()"
                    )
                else:
                    print(
                        "PublishScalarsService:46: TRY: publish([publish_object.data])"
                    )
                    result = publish(
                        [publish_object.data], node.acc, msg.sigma, user_key=verify_key
                    )
                    print(
                        "PublishScalarsService:50: SUCCESS: publish([publish_object.data])"
                    )
                results.append(result)
            except Exception as e:
                print("PublishScalarsService:53: EXCEPTION - missing id")
                log = (
                    f"Unable to Get Object with ID {publish_id} from store. "
                    + f"Possible dangling Pointer. {e}"
                )
                traceback_and_raise(Exception(log))

        # give the caller permission to download this
        read_permissions: TypeDict[VerifyKey, UID] = {verify_key: None}
        search_permissions: TypeDict[VerifyKey, Optional[UID]] = {verify_key: None}

        if len(results) == 1:
            results = results[0]
        print("PublishScalarsService:67: storable = StorableObject(")
        storable = StorableObject(
            id=msg.id_at_location,
            data=results,
            description=f"Approved AutoDP Result: {msg.id_at_location}",
            read_permissions=read_permissions,
            search_permissions=search_permissions,
        )
        print(
            "PublishScalarsService:74:  msg.id_at_location == "
            + str(msg.id_at_location)
            + " obj_type:"
            + str(type(storable))
        )
        print(
            "PublishScalarsService:75: TRY: node.store[msg.id_at_location] = storable"
        )
        node.store[msg.id_at_location] = storable
        print(
            "PublishScalarsService:77: SUCCESS: node.store[msg.id_at_location] = storable"
        )

    @staticmethod
    def message_handler_types() -> TypeList[Type[PublishScalarsAction]]:
        return [PublishScalarsAction]



# stdlib
from typing import Any
from typing import Dict

# relative
from .....experimental_flags import flags
from .....logger import logger
from ...abstract.node import AbstractNodeClient
from ...domain.enums import ResponseObjectEnum
from ..exceptions import AuthorizationError
from ..node_service.user_auth.user_auth_messages import UserLoginMessageWithReply

if flags.USE_NEW_SERVICE:
    # relative
    from ..node_service.user_manager.new_user_messages import CreateUserMessage
    from ..node_service.user_manager.new_user_messages import DeleteUserMessage
    from ..node_service.user_manager.new_user_messages import GetUserMessage
    from ..node_service.user_manager.new_user_messages import GetUsersMessage
    from ..node_service.user_manager.new_user_messages import UpdateUserMessage
else:
    # relative
    # type: ignore[override]
    from ..node_service.user_manager.user_messages import CreateUserMessage  # type: ignore
    from ..node_service.user_manager.user_messages import DeleteUserMessage  # type: ignore
    from ..node_service.user_manager.user_messages import GetUserMessage  # type: ignore
    from ..node_service.user_manager.user_messages import GetUsersMessage  # type: ignore
    from ..node_service.user_manager.user_messages import UpdateUserMessage  # type: ignore

# relative
from .request_api import RequestAPI


class UserRequestAPI(RequestAPI):
    def __init__(self, client: AbstractNodeClient):
        super().__init__(
            client=client,
            create_msg=CreateUserMessage,
            get_msg=GetUserMessage,
            get_all_msg=GetUsersMessage,
            update_msg=UpdateUserMessage,
            delete_msg=DeleteUserMessage,
            response_key=ResponseObjectEnum.USER,
        )

    def __getitem__(self, key: int) -> Any:
        return self.get(user_id=key)

    def __delitem__(self, key: int) -> None:
        self.delete(user_id=key)

    def create(self, **kwargs: Any) -> None:
        try:
            if "pdf" in kwargs.keys():
                response = self.client.routes[0].connection.send_files(  # type: ignore
                    "/users",
                    kwargs.get("pdf"),
                    form_name="new_user",
                    form_values=kwargs,  # type: ignore
                )  # type: ignore
                logger.info(response)
            else:
                if flags.USE_NEW_SERVICE:
                    response = self.perform_request(
                        syft_msg=self._create_message, content=kwargs  # type: ignore
                    )
                    logger.info(response.message)
                else:
                    response = self.perform_api_request(
                        syft_msg=self._create_message, content=kwargs
                    )
                    logger.info(response.resp_msg)

        except Exception as e:
            print("failing to create user", e)
            try:
                for user in self.all():
                    if user["email"] == kwargs["email"]:
                        print(
                            "Ignoring: user with email:"
                            + user["email"]
                            + " already exists"
                        )
                        return
            except AuthorizationError as exc:
                print("No permission to check users", exc)

            raise e

    def login(self, email: str, password: str) -> Dict[str, Any]:
        response = self.perform_api_request_generic(
            syft_msg=UserLoginMessageWithReply,
            content={"email": email, "password": password},
        )

        return response.payload.kwargs.upcast()  # type: ignore



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/adp/scalar_manager.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.lib.python import dict_pb2 as proto_dot_lib_dot_python_dot_dict__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n#proto/core/adp/scalar_manager.proto\x12\rsyft.core.adp\x1a\x1bproto/lib/python/dict.proto"\x0e\n\x0cPrimeFactory"\x85\x01\n"VirtualMachinePrivateScalarManager\x12\x32\n\rprime_factory\x18\x01 \x01(\x0b\x32\x1b.syft.core.adp.PrimeFactory\x12+\n\x0cprime2symbol\x18\x02 \x01(\x0b\x32\x15.syft.lib.python.Dictb\x06proto3'
)


_PRIMEFACTORY = DESCRIPTOR.message_types_by_name["PrimeFactory"]
_VIRTUALMACHINEPRIVATESCALARMANAGER = DESCRIPTOR.message_types_by_name[
    "VirtualMachinePrivateScalarManager"
]
PrimeFactory = _reflection.GeneratedProtocolMessageType(
    "PrimeFactory",
    (_message.Message,),
    {
        "DESCRIPTOR": _PRIMEFACTORY,
        "__module__": "proto.core.adp.scalar_manager_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.adp.PrimeFactory)
    },
)
_sym_db.RegisterMessage(PrimeFactory)

VirtualMachinePrivateScalarManager = _reflection.GeneratedProtocolMessageType(
    "VirtualMachinePrivateScalarManager",
    (_message.Message,),
    {
        "DESCRIPTOR": _VIRTUALMACHINEPRIVATESCALARMANAGER,
        "__module__": "proto.core.adp.scalar_manager_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.adp.VirtualMachinePrivateScalarManager)
    },
)
_sym_db.RegisterMessage(VirtualMachinePrivateScalarManager)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _PRIMEFACTORY._serialized_start = 83
    _PRIMEFACTORY._serialized_end = 97
    _VIRTUALMACHINEPRIVATESCALARMANAGER._serialized_start = 100
    _VIRTUALMACHINEPRIVATESCALARMANAGER._serialized_end = 233
# @@protoc_insertion_point(module_scope)



# relative
from ..abstract.node import AbstractNode
from ..common.node_manager.node_manager import NodeManager
from ..common.node_manager.node_route_manager import NodeRouteManager
from ..common.node_manager.role_manager import RoleManager
from ..common.node_manager.user_manager import UserManager


class NodeServiceInterface(AbstractNode):
    node: NodeManager
    node_route: NodeRouteManager
    users: UserManager
    roles: RoleManager



# third party
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String

# relative
from . import Base


class NodeRoute(Base):
    __tablename__ = "node_route"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    node_id = Column(Integer, ForeignKey("node.id"))
    host_or_ip = Column(String(255))
    is_vpn = Column(Boolean(), default=False)



# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from ......util import traceback_and_raise
from ....abstract.node import AbstractNode
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from .get_remaining_budget_messages import GetRemainingBudgetMessage
from .get_remaining_budget_messages import GetRemainingBudgetReplyMessage


class GetRemainingBudgetService(ImmediateNodeServiceWithReply):
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: AbstractNode,
        msg: GetRemainingBudgetMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> GetRemainingBudgetReplyMessage:

        if verify_key is None:
            traceback_and_raise(
                "Can't process GetRemainingBudgetService with no verification key."
            )

        try:

            result = node.acc.get_remaining_budget(  # type: ignore
                user_key=verify_key, returned_epsilon_is_private=False
            )

            return GetRemainingBudgetReplyMessage(budget=result, address=msg.address)
        except Exception as e:
            log = (
                f"Unable to get remaining budget for {verify_key}. "
                + f"Possible dangling Pointer. {e}"
            )
            traceback_and_raise(Exception(log))

    @staticmethod
    def message_handler_types() -> List[Type[GetRemainingBudgetMessage]]:
        return [GetRemainingBudgetMessage]



# stdlib
import re
from typing import Any

# third party
from google.protobuf.message import Message

# relative
from ....logger import traceback_and_raise
from ....proto.util.data_message_pb2 import DataMessage
from ....util import index_syft_by_module_name
from .capnp import CAPNP_END_MAGIC_HEADER_BYTES
from .capnp import CAPNP_REGISTRY
from .capnp import CAPNP_START_MAGIC_HEADER
from .capnp import CAPNP_START_MAGIC_HEADER_BYTES
from .types import Deserializeable

PROTOBUF_START_MAGIC_HEADER = "protobuf:"
PROTOBUF_START_MAGIC_HEADER_BYTES = PROTOBUF_START_MAGIC_HEADER.encode("utf-8")


# WARNING: This code has more 🐉 Dragons than a game of D&D 🗡🧙🎲
# you were warned...
# enter at your own peril...
# seriously, get some 🧪 HP Potions and 📜 TP Scrolls ready...
def _deserialize(
    blob: Deserializeable,
    from_proto: bool = True,
    from_bytes: bool = False,
) -> Any:
    """We assume you're deserializing a protobuf object by default

    This function deserializes from encoding to a Python object. There are a few ways of
    using this function:
    1. An Message object is passed,q this will transform a protobuf message into its associated
    class.
    the from_proto has to be set (it is by default).
    2. Bytes are passed. This requires the from_bytes flag set the schema_type specified.
    We cannot (and we should not) be able to get the schema_type from the binary representation.

    Note: The only format that does not require the schema_type is when we are passing
    Messages directly.

    Raises: ValueError if you are not setting one from_<protocol> flag.
            ValueError if you are deserializing a data type that requires a schema type and not
            providing one.
            TypeError if you are trying to deserialize an unsupported type.

    :param blob: this parameter is the data to be deserialized from various formats.
    :type blob: Union[str, dict, bytes, Messages]
    :param from_proto: set this flag to True if you want to deserialize a protobuf message.
    :param from_bytes: set this flag to True if you want to deserialize a binary object.
    :type from_bytes: bool
    :return: a deserialized form of the object on which _deserialize() is called.
    :rtype: Serializable
    """

    deserialization_error = TypeError(
        "You tried to deserialize an unsupported type. This can be caused by "
        "several reasons. Either you are actively writing Syft code and forgot "
        "to create one, or you are trying to deserialize an object which was "
        "serialized using a different version of Syft and the object you tried "
        "to deserialize is not supported in this version."
    )

    # try to decode capnp first
    if isinstance(blob, bytes):
        try:
            return deserialize_capnp(buf=blob)
        except CapnpMagicBytesNotFound:  # nosec
            # probably not capnp bytes
            pass
        except Exception as e:
            # capnp magic bytes found but another problem has occured
            print("failed capnp deserialize", e)
            raise e

    if from_bytes:
        data_message = DataMessage()
        data_message.ParseFromString(blob)
        obj_type = index_syft_by_module_name(fully_qualified_name=data_message.obj_type)
        get_protobuf_schema = getattr(obj_type, "get_protobuf_schema", None)

        if not callable(get_protobuf_schema):
            traceback_and_raise(deserialization_error)

        protobuf_type = get_protobuf_schema()
        blob = protobuf_type()

        if not isinstance(blob, Message):
            traceback_and_raise(deserialization_error)

        blob.ParseFromString(data_message.content)

    # lets try to lookup the type we are deserializing
    # TODO: This needs to be cleaned up in GenerateWrapper and made more consistent.
    # There are serveral code paths that come through here and use different ways to
    # match and overload protobuf -> deserialize type
    obj_type = getattr(type(blob), "schema2type", None)
    # relative
    from .recursive import rs_get_protobuf_schema
    from .recursive import rs_proto2object

    if obj_type is None:
        # TODO: This can probably be removed now we have lists of obj_types
        obj_type = getattr(blob, "obj_type", None)
        if isinstance(blob, rs_get_protobuf_schema()):
            res = rs_proto2object(proto=blob)
            if getattr(res, "temporary_box", False) and hasattr(res, "upcast"):
                return res.upcast()
            return res

        if obj_type is None:
            traceback_and_raise(deserialization_error)

        obj_type = index_syft_by_module_name(fully_qualified_name=obj_type)  # type: ignore
        obj_type = getattr(obj_type, "_sy_serializable_wrapper_type", obj_type)
    elif isinstance(obj_type, list):
        if isinstance(blob, rs_get_protobuf_schema()):
            res = rs_proto2object(proto=blob)
            if getattr(res, "temporary_box", False) and hasattr(res, "upcast"):
                return res.upcast()
            return res
        elif len(obj_type) == 1:
            obj_type = obj_type[0]
        else:
            # this means we have multiple classes that use the same proto but use the
            # obj_type field to differentiate, so lets figure out which one in the list
            obj_type_re = r'obj_type: "(.+)"'
            # the first obj_type in the protobuf will be the main outer type
            obj_types = re.findall(obj_type_re, str(blob))
            if len(obj_types) > 0:
                real_obj_type = obj_types[0]
                for possible_type in obj_type:
                    possible_type_match = possible_type
                    if hasattr(possible_type, "wrapped_type"):
                        possible_type_match = possible_type.wrapped_type()
                    # get the str inside <class ...>, fqn in sympy is different
                    real_obj_type_str = str(possible_type_match).split("'")[1]
                    if real_obj_type.endswith("Wrapper"):
                        real_obj_type = real_obj_type[:-7]  # remove the last Wrapper

                    if real_obj_type == real_obj_type_str or real_obj_type.endswith(
                        real_obj_type_str
                    ):
                        # found it, lets overwrite obj_type and break
                        obj_type = possible_type
                        break

    if not isinstance(obj_type, type):
        traceback_and_raise(f"{deserialization_error}. {type(blob)}")

    _proto2object = getattr(obj_type, "_proto2object", None)
    if not callable(_proto2object):
        traceback_and_raise(deserialization_error)

    res = _proto2object(proto=blob)

    # if its a temporary_box upcast
    if getattr(res, "temporary_box", False) and hasattr(res, "upcast"):
        return res.upcast()

    return res


class CapnpMagicBytesNotFound(Exception):
    pass


def deserialize_capnp(buf: bytes) -> Any:
    # only search 100 bytes to prevent wasting time on large files
    search_range = 100
    header_bytes = buf[0:search_range]
    chars = bytearray()
    # filter header bytes
    for i in header_bytes:
        # only allow ascii letters or : in headers and class name to prevent lookup
        # breaking somehow, when packing weird stuff like \x03 ends up in the string
        # e.g. NDimEntityPhiTensor -> ND\x03imEntityPhiTensor
        if i in range(65, 91) or i in range(97, 123) or i == 58:
            chars.append(i)
    header_bytes = bytes(chars)

    proto_start_index = header_bytes.find(PROTOBUF_START_MAGIC_HEADER_BYTES)
    start_index = header_bytes.find(CAPNP_START_MAGIC_HEADER_BYTES)
    if proto_start_index != -1 and (proto_start_index < start_index):
        # we have protobuf on the outside
        raise CapnpMagicBytesNotFound(
            f"protobuf Magic Header {PROTOBUF_START_MAGIC_HEADER} found in bytes"
        )
    if start_index == -1:
        raise CapnpMagicBytesNotFound(
            f"capnp Magic Header {CAPNP_START_MAGIC_HEADER}" + "not found in bytes"
        )
    start_index += len(CAPNP_START_MAGIC_HEADER_BYTES)
    end_index = header_bytes.index(CAPNP_END_MAGIC_HEADER_BYTES)
    class_name_bytes = header_bytes[start_index:end_index]

    class_name = class_name_bytes.decode("utf-8")
    if class_name not in CAPNP_REGISTRY:
        raise Exception(
            f"Found capnp Magic Header: {CAPNP_START_MAGIC_HEADER} "
            + f"and Class {class_name} but no mapping in capnp registry {CAPNP_REGISTRY}"
        )
    return CAPNP_REGISTRY[class_name](buf=buf)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/user_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.lib.python import dict_pb2 as proto_dot_lib_dot_python_dot_dict__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\'proto/grid/messages/user_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto\x1a\x1bproto/lib/python/dict.proto"\x8f\x02\n\x11\x43reateUserMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\r\n\x05\x65mail\x18\x03 \x01(\t\x12\x10\n\x08password\x18\x04 \x01(\t\x12\x0c\n\x04role\x18\x05 \x01(\t\x12\x0c\n\x04name\x18\x06 \x01(\t\x12\x0f\n\x07website\x18\t \x01(\t\x12\x13\n\x0binstitution\x18\n \x01(\t\x12\x0f\n\x07\x64\x61\x61_pdf\x18\x0b \x01(\x0c\x12\'\n\x08reply_to\x18\x0c \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0e\n\x06\x62udget\x18\r \x01(\x01"\x99\x01\n\x0eGetUserMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07user_id\x18\x03 \x01(\x05\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\xbb\x01\n\x1bProcessUserCandidateMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x14\n\x0c\x63\x61ndidate_id\x18\x02 \x01(\x05\x12\x0e\n\x06status\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Address"\x88\x01\n\x0fGetUserResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x63ontent\x18\x02 \x01(\x0b\x32\x15.syft.lib.python.Dict\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\x8e\x01\n\x14GetCandidatesMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8e\x01\n\x15GetCandidatesResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x63ontent\x18\x02 \x03(\x0b\x32\x15.syft.lib.python.Dict\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\x89\x01\n\x0fGetUsersMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x89\x01\n\x10GetUsersResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x63ontent\x18\x02 \x03(\x0b\x32\x15.syft.lib.python.Dict\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\x9c\x01\n\x11\x44\x65leteUserMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07user_id\x18\x03 \x01(\x05\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9f\x02\n\x11UpdateUserMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07user_id\x18\x03 \x01(\x05\x12\r\n\x05\x65mail\x18\x04 \x01(\t\x12\x10\n\x08password\x18\x05 \x01(\t\x12\x0c\n\x04role\x18\x06 \x01(\t\x12\x0e\n\x06groups\x18\x07 \x01(\t\x12\x0c\n\x04name\x18\x08 \x01(\t\x12\x0e\n\x06\x62udget\x18\t \x01(\x01\x12\x13\n\x0binstitution\x18\n \x01(\t\x12\x0f\n\x07website\x18\x0b \x01(\t\x12\'\n\x08reply_to\x18\x0c \x01(\x0b\x32\x15.syft.core.io.Address"\xc7\x01\n\x12SearchUsersMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\r\n\x05\x65mail\x18\x03 \x01(\t\x12\x0c\n\x04role\x18\x04 \x01(\t\x12\x0e\n\x06groups\x18\x05 \x01(\t\x12\x0c\n\x04name\x18\x06 \x01(\t\x12\'\n\x08reply_to\x18\x07 \x01(\x0b\x32\x15.syft.core.io.Address"\x8c\x01\n\x13SearchUsersResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x63ontent\x18\x02 \x03(\x0b\x32\x15.syft.lib.python.Dict\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)


_CREATEUSERMESSAGE = DESCRIPTOR.message_types_by_name["CreateUserMessage"]
_GETUSERMESSAGE = DESCRIPTOR.message_types_by_name["GetUserMessage"]
_PROCESSUSERCANDIDATEMESSAGE = DESCRIPTOR.message_types_by_name[
    "ProcessUserCandidateMessage"
]
_GETUSERRESPONSE = DESCRIPTOR.message_types_by_name["GetUserResponse"]
_GETCANDIDATESMESSAGE = DESCRIPTOR.message_types_by_name["GetCandidatesMessage"]
_GETCANDIDATESRESPONSE = DESCRIPTOR.message_types_by_name["GetCandidatesResponse"]
_GETUSERSMESSAGE = DESCRIPTOR.message_types_by_name["GetUsersMessage"]
_GETUSERSRESPONSE = DESCRIPTOR.message_types_by_name["GetUsersResponse"]
_DELETEUSERMESSAGE = DESCRIPTOR.message_types_by_name["DeleteUserMessage"]
_UPDATEUSERMESSAGE = DESCRIPTOR.message_types_by_name["UpdateUserMessage"]
_SEARCHUSERSMESSAGE = DESCRIPTOR.message_types_by_name["SearchUsersMessage"]
_SEARCHUSERSRESPONSE = DESCRIPTOR.message_types_by_name["SearchUsersResponse"]
CreateUserMessage = _reflection.GeneratedProtocolMessageType(
    "CreateUserMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEUSERMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.CreateUserMessage)
    },
)
_sym_db.RegisterMessage(CreateUserMessage)

GetUserMessage = _reflection.GeneratedProtocolMessageType(
    "GetUserMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETUSERMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetUserMessage)
    },
)
_sym_db.RegisterMessage(GetUserMessage)

ProcessUserCandidateMessage = _reflection.GeneratedProtocolMessageType(
    "ProcessUserCandidateMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _PROCESSUSERCANDIDATEMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.ProcessUserCandidateMessage)
    },
)
_sym_db.RegisterMessage(ProcessUserCandidateMessage)

GetUserResponse = _reflection.GeneratedProtocolMessageType(
    "GetUserResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETUSERRESPONSE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetUserResponse)
    },
)
_sym_db.RegisterMessage(GetUserResponse)

GetCandidatesMessage = _reflection.GeneratedProtocolMessageType(
    "GetCandidatesMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETCANDIDATESMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetCandidatesMessage)
    },
)
_sym_db.RegisterMessage(GetCandidatesMessage)

GetCandidatesResponse = _reflection.GeneratedProtocolMessageType(
    "GetCandidatesResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETCANDIDATESRESPONSE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetCandidatesResponse)
    },
)
_sym_db.RegisterMessage(GetCandidatesResponse)

GetUsersMessage = _reflection.GeneratedProtocolMessageType(
    "GetUsersMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETUSERSMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetUsersMessage)
    },
)
_sym_db.RegisterMessage(GetUsersMessage)

GetUsersResponse = _reflection.GeneratedProtocolMessageType(
    "GetUsersResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETUSERSRESPONSE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetUsersResponse)
    },
)
_sym_db.RegisterMessage(GetUsersResponse)

DeleteUserMessage = _reflection.GeneratedProtocolMessageType(
    "DeleteUserMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _DELETEUSERMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.DeleteUserMessage)
    },
)
_sym_db.RegisterMessage(DeleteUserMessage)

UpdateUserMessage = _reflection.GeneratedProtocolMessageType(
    "UpdateUserMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATEUSERMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.UpdateUserMessage)
    },
)
_sym_db.RegisterMessage(UpdateUserMessage)

SearchUsersMessage = _reflection.GeneratedProtocolMessageType(
    "SearchUsersMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _SEARCHUSERSMESSAGE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.SearchUsersMessage)
    },
)
_sym_db.RegisterMessage(SearchUsersMessage)

SearchUsersResponse = _reflection.GeneratedProtocolMessageType(
    "SearchUsersResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _SEARCHUSERSRESPONSE,
        "__module__": "proto.grid.messages.user_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.SearchUsersResponse)
    },
)
_sym_db.RegisterMessage(SearchUsersResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _CREATEUSERMESSAGE._serialized_start = 161
    _CREATEUSERMESSAGE._serialized_end = 432
    _GETUSERMESSAGE._serialized_start = 435
    _GETUSERMESSAGE._serialized_end = 588
    _PROCESSUSERCANDIDATEMESSAGE._serialized_start = 591
    _PROCESSUSERCANDIDATEMESSAGE._serialized_end = 778
    _GETUSERRESPONSE._serialized_start = 781
    _GETUSERRESPONSE._serialized_end = 917
    _GETCANDIDATESMESSAGE._serialized_start = 920
    _GETCANDIDATESMESSAGE._serialized_end = 1062
    _GETCANDIDATESRESPONSE._serialized_start = 1065
    _GETCANDIDATESRESPONSE._serialized_end = 1207
    _GETUSERSMESSAGE._serialized_start = 1210
    _GETUSERSMESSAGE._serialized_end = 1347
    _GETUSERSRESPONSE._serialized_start = 1350
    _GETUSERSRESPONSE._serialized_end = 1487
    _DELETEUSERMESSAGE._serialized_start = 1490
    _DELETEUSERMESSAGE._serialized_end = 1646
    _UPDATEUSERMESSAGE._serialized_start = 1649
    _UPDATEUSERMESSAGE._serialized_end = 1936
    _SEARCHUSERSMESSAGE._serialized_start = 1939
    _SEARCHUSERSMESSAGE._serialized_end = 2138
    _SEARCHUSERSRESPONSE._serialized_start = 2141
    _SEARCHUSERSRESPONSE._serialized_end = 2281
# @@protoc_insertion_point(module_scope)



# stdlib
from copy import deepcopy
import functools
import operator
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from uuid import UUID

# third party
import numpy as np

# relative
from ....common.uid import UID
from ....store.storeable_object import StorableObject
from ....tensor.smpc import utils
from ....tensor.smpc.share_tensor import ShareTensor
from ...abstract.node import AbstractNode
from .beaver_action import BeaverAction
from .greenlets_switch import beaver_retrieve_object
from .smpc_action_message import SMPCActionMessage
from .smpc_action_seq_batch_message import SMPCActionSeqBatchMessage


def get_action_generator_from_op(
    operation_str: str, nr_parties: int
) -> Callable[[UID, UID, int, Any], Any]:
    """ "
    Get the generator for the operation provided by the argument
    Arguments:
        operation_str (str): the name of the operation

    """
    return functools.partial(MAP_FUNC_TO_ACTION[operation_str], nr_parties)


def get_id_at_location_from_op(seed: bytes, operation_str: str) -> UID:
    generator = np.random.default_rng(seed)
    return UID(UUID(bytes=generator.bytes(16)))


# TODO: node.address is not really ,should be removed later.


def smpc_basic_op(
    op_str: str,
    nr_parties: int,
    self_id: UID,
    other_id: UID,
    seed_id_locations: int,
    node: AbstractNode,
) -> List[SMPCActionMessage]:
    # relative
    from ..... import Tensor

    """Generator for SMPC public/private operations add/sub"""

    generator = np.random.default_rng(seed_id_locations)
    result_id = UID(UUID(bytes=generator.bytes(16)))
    other = node.store.get(other_id).data

    actions = []
    if isinstance(other, (ShareTensor, Tensor)):
        # All parties should add the other share if empty list
        actions.append(
            SMPCActionMessage(
                f"mpc_{op_str}",
                self_id=self_id,
                args_id=[other_id],
                kwargs_id={},
                ranks_to_run_action=list(range(nr_parties)),
                result_id=result_id,
                address=node.address,
            )
        )
    else:
        actions.append(
            SMPCActionMessage(
                "mpc_noop",
                self_id=self_id,
                args_id=[],
                kwargs_id={},
                ranks_to_run_action=list(range(1, nr_parties)),
                result_id=result_id,
                address=node.address,
            )
        )

        # Only rank 0 (the first party) would do the add/sub for the public value
        actions.append(
            SMPCActionMessage(
                f"mpc_{op_str}",
                self_id=self_id,
                args_id=[other_id],
                kwargs_id={},
                ranks_to_run_action=[0],
                result_id=result_id,
                address=node.address,
            )
        )

    return actions


# TODO : Should move to spdz directly in syft/core/smpc
def spdz_multiply(
    x: ShareTensor,
    y: ShareTensor,
    eps_id: UID,
    delta_id: UID,
    a_share: ShareTensor,
    b_share: ShareTensor,
    c_share: ShareTensor,
    node: Optional[Any] = None,
) -> ShareTensor:
    # relative
    from ..... import Tensor

    TENSOR_FLAG = False
    if isinstance(x, Tensor) and isinstance(y, Tensor):
        TENSOR_FLAG = True
        t1 = x
        t2 = y
        x = x.child.child
        y = y.child.child

    nr_parties = x.nr_parties

    ring_size = utils.get_ring_size(x.ring_size, y.ring_size)

    eps = beaver_retrieve_object(node, eps_id, nr_parties)  # type: ignore
    delta = beaver_retrieve_object(node, delta_id, nr_parties)  # type: ignore

    # TODO : Should refactor fixed precision tensor later

    eps = sum(eps.data)  # type: ignore
    delta = sum(delta.data)  # type:ignore

    eps_b = eps * b_share.child
    delta_a = delta * a_share.child  # Symmetric only for mul

    tensor = c_share + eps_b + delta_a
    if x.rank == 0:
        mul_op = ShareTensor.get_op(ring_size, "mul")
        eps_delta = mul_op(eps.child, delta.child)  # type: ignore
        tensor = tensor + eps_delta

    share = x.copy_tensor()
    share.child = tensor.child  # As we do not use fixed point we neglect truncation.

    if TENSOR_FLAG:
        t = t1 * t2
        t.child.child = share
        return t
    else:
        return share


# TODO : Should move to spdz directly in syft/core/smpc
def spdz_mask(
    x: ShareTensor,
    y: ShareTensor,
    eps_id: UID,
    delta_id: UID,
    a_share: ShareTensor,
    b_share: ShareTensor,
    c_share: ShareTensor,
    node: Optional[AbstractNode] = None,
) -> None:
    # relative
    from ..... import Tensor

    if isinstance(x, Tensor) and isinstance(y, Tensor):
        x = x.child.child
        y = y.child.child

    if node is None:
        raise ValueError("Node context should be passed to spdz mask")

    clients = ShareTensor.login_clients(x.parties_info)

    eps = x - a_share  # beaver intermediate values.
    delta = y - b_share

    client_id_map = {client.id: client for client in clients}
    curr_client = client_id_map[node.id]  # type: ignore
    beaver_action = BeaverAction(
        eps=eps,
        eps_id=eps_id,
        delta=delta,
        delta_id=delta_id,
        address=curr_client.address,
    )
    beaver_action.execute_action(node, None)

    for _, client in enumerate(clients):
        if client != curr_client:
            beaver_action.address = client.address
            client.send_immediate_msg_without_reply(msg=beaver_action)


def smpc_mul(
    nr_parties: int,
    self_id: UID,
    other_id: UID,
    a_shape_id: Optional[UID] = None,
    b_shape_id: Optional[UID] = None,
    seed_id_locations: Optional[int] = None,
    node: Optional[AbstractNode] = None,
) -> SMPCActionSeqBatchMessage:
    """Generator for the smpc_mul with a public value"""
    # relative
    from ..... import Tensor

    if seed_id_locations is None or node is None:
        raise ValueError(
            f"The values seed_id_locations{seed_id_locations}, Node:{node} should not be None"
        )
    generator = np.random.default_rng(seed_id_locations)
    result_id = UID(UUID(bytes=generator.bytes(16)))
    other = node.store.get(other_id).data

    actions = []
    if isinstance(other, (ShareTensor, Tensor)) and (a_shape_id and b_shape_id):
        # crypto_store = ShareTensor.crypto_store
        # _self = node.store.get(self_id).data
        # a_share, b_share, c_share = crypto_store.get_primitives_from_store("beaver_mul", _self.shape, other.shape)
        if isinstance(other, ShareTensor):
            ring_size = other.ring_size
        else:
            ring_size = other.child.child.ring_size

        mask_result = UID(UUID(bytes=generator.bytes(16)))
        eps_id = UID(UUID(bytes=generator.bytes(16)))
        delta_id = UID(UUID(bytes=generator.bytes(16)))
        a_shape = node.store.get(a_shape_id).data  # type: ignore
        b_shape = node.store.get(b_shape_id).data  # type: ignore
        crypto_store = ShareTensor.crypto_store
        a_share, b_share, c_share = crypto_store.get_primitives_from_store(
            "beaver_mul", a_shape=a_shape, b_shape=b_shape, ring_size=ring_size, remove=True  # type: ignore
        )

        actions.append(
            SMPCActionMessage(
                "spdz_mask",
                self_id=self_id,
                args_id=[other_id],
                kwargs_id={},
                kwargs={
                    "eps_id": eps_id,
                    "delta_id": delta_id,
                    "a_share": a_share,
                    "b_share": b_share,
                    "c_share": c_share,
                },
                ranks_to_run_action=list(range(nr_parties)),
                result_id=mask_result,
                address=node.address,
            )
        )

        actions.append(
            SMPCActionMessage(
                "spdz_multiply",
                self_id=self_id,
                args_id=[other_id],
                kwargs_id={},
                kwargs={
                    "eps_id": eps_id,
                    "delta_id": delta_id,
                    "a_share": a_share,
                    "b_share": b_share,
                    "c_share": c_share,
                },
                ranks_to_run_action=list(range(nr_parties)),
                result_id=result_id,
                address=node.address,
            )
        )

    else:
        # All ranks should multiply by that public value
        actions.append(
            SMPCActionMessage(
                "mpc_mul",
                self_id=self_id,
                args_id=[other_id],
                kwargs_id={},
                ranks_to_run_action=list(range(nr_parties)),
                result_id=result_id,
                address=node.address,
            )
        )
    batch = SMPCActionSeqBatchMessage(smpc_actions=actions, address=node.address)
    return batch


def local_decomposition(
    x: ShareTensor,
    ring_size: int,
    bitwise: bool,
    seed_id_locations: str,
    node: Any,
    read_permissions: Dict[Any, Any],
) -> None:
    """Performs local decomposition to generate shares of shares.

    Args:
        x (ShareTensor) : input ShareTensor.
        ring_size (str) : Ring size to generate decomposed shares in.
        bitwise (bool): Perform bit level decomposition on bits if set.

    Returns:
        List[List[ShareTensor]]: Decomposed shares in the given ring size.
    """
    # Currently we cannot serialize integers exceeding signed positive integer
    # TODO: can be modified to google.protobuf.any
    seed_id_locations = int(seed_id_locations)  # type: ignore
    generator = np.random.default_rng(seed_id_locations)
    # Skip the first ID ,as it is used for None return type in run class method.
    _ = UID(UUID(bytes=generator.bytes(16)))
    # relative
    from ..... import Tensor

    TENSOR_FLAG = False
    if isinstance(x, Tensor):
        TENSOR_FLAG = True
        t = x
        x = x.child.child

    rank = x.rank
    nr_parties = x.nr_parties
    numpy_type = utils.RING_SIZE_TO_TYPE[ring_size]
    shape = x.shape
    zero = np.zeros(shape, numpy_type)

    input_shares = []

    if bitwise:
        ring_bits = utils.get_nr_bits(x.ring_size)  # for bit-wise decomposition
        input_shares = [x.bit_extraction(idx) for idx in range(ring_bits)]
    else:
        input_shares.append(x)

    for share in input_shares:

        for i in range(nr_parties):
            id_at_location = UID(UUID(bytes=generator.bytes(16)))
            sh = x.copy_tensor()
            sh.ring_size = ring_size
            if rank != i:
                sh.child = deepcopy(zero)
            else:
                sh.child = deepcopy(share.child.astype(numpy_type))
            if TENSOR_FLAG:
                t_sh = t.copy()
                t_sh.child.child = sh
                data = t_sh
            else:
                data = sh
            store_obj = StorableObject(
                id=id_at_location,
                data=data,
                read_permissions=read_permissions,
            )
            node.store[id_at_location] = store_obj

    return None


def bit_decomposition(
    nr_parties: int,
    self_id: UID,
    ring_size: UID,
    bitwise: UID,
    seed_id_locations: int,
    node: Any,
) -> List[SMPCActionMessage]:
    generator = np.random.default_rng(seed_id_locations)
    result_id = UID(UUID(bytes=generator.bytes(16)))

    actions = []

    actions.append(
        SMPCActionMessage(
            "local_decomposition",
            self_id=self_id,
            args_id=[ring_size, bitwise],
            kwargs_id={},
            kwargs={"seed_id_locations": str(seed_id_locations)},
            ranks_to_run_action=list(range(nr_parties)),
            result_id=result_id,
            address=node.address,
        )
    )

    return actions


# Given an SMPC Action map it to an action constructor
MAP_FUNC_TO_ACTION: Dict[
    str, Callable[[int, UID, UID, int, Any], List[SMPCActionMessage]]
] = {
    "__add__": functools.partial(smpc_basic_op, "add"),
    "__sub__": functools.partial(smpc_basic_op, "sub"),
    "__mul__": smpc_mul,  # type: ignore
    "bit_decomposition": bit_decomposition,  # type: ignore
    # "__gt__": smpc_gt,  # type: ignore TODO: this should be added back when we have only one action
}


# Map given an action map it to a function that should be run on the shares"
_MAP_ACTION_TO_FUNCTION: Dict[str, Callable[..., Any]] = {
    "mpc_add": operator.add,
    "mpc_sub": operator.sub,
    "mpc_mul": operator.mul,
    "spdz_mask": spdz_mask,
    "spdz_multiply": spdz_multiply,
    "local_decomposition": local_decomposition,
    "mpc_noop": deepcopy,
}


#
# def smpc_gt(
#     nr_parties: int,
#     self_id: UID,
#     other_id: UID,
#     seed_id_locations: int,
#     node: Any,
#     client: Any,
# ) -> List[SMPCActionMessage]:
#     """Generator for the smpc_mul with a public value"""
#     generator = np.random.default_rng(seed_id_locations)

#     result_id = UID(UUID(bytes=generator.bytes(16)))
#     sub_result = UID(UUID(bytes=generator.bytes(16)))

#     x = node.store.get(self_id).data  # noqa
#     y = node.store.get(other_id).data

#     if not isinstance(y, ShareTensor):
#         raise ValueError("Only private compare works at the moment")

#     actions = []
#     actions.append(
#         SMPCActionMessage(
#             "mpc_sub",
#             self_id=self_id,
#             args_id=[other_id],
#             kwargs_id={},
#             ranks_to_run_action=list(range(nr_parties)),
#             result_id=sub_result,
#             address=client.address,
#         )
#     )

#     actions.append(
#         SMPCActionMessage(
#             "bit_decomposition",
#             self_id=sub_result,
#             args_id=[],
#             # TODO: This value needs to be changed to something else and probably used
#             # directly the przs_generator from ShareTensor - check bit_decomposition function
#             kwargs_id={},
#             ranks_to_run_action=list(range(nr_parties)),
#             result_id=result_id,
#             address=client.address,
#         )
#     )
#     return actions



# third party
from pydantic import BaseSettings

# relative
from ...adp.adversarial_accountant import AdversarialAccountant
from ..abstract.node_service_interface import NodeServiceInterface
from ..common.node_manager.association_request_manager import AssociationRequestManager
from ..common.node_manager.dataset_manager import DatasetManager
from ..common.node_manager.role_manager import RoleManager
from ..common.node_manager.setup_manager import SetupManager
from ..common.node_manager.user_manager import UserManager


class DomainInterface(NodeServiceInterface):
    users: UserManager
    roles: RoleManager
    association_requests: AssociationRequestManager
    datasets: DatasetManager
    setup: SetupManager
    acc: AdversarialAccountant
    settings: BaseSettings



# stdlib
from enum import Enum
from typing import Any

# relative
from ..common.serde.serializable import serializable

global_value: int = 5


def global_function() -> int:
    return global_value


@serializable(recursive_serde=True)
class A:
    """This is a test class to be used in functionality test."""

    __slots__ = ["_private_attr", "n"]
    __attr_allowlist__ = ["_private_attr", "n"]

    static_attr: int = 4

    def __init__(self) -> None:
        self._private_attr: float = 5.5

    def __len__(self) -> int:
        return 3

    def __iter__(self) -> "A":
        self.n = 1
        return self

    def __next__(self) -> int:
        if self.n < A.static_attr:
            result = self.n
            self.n += 1
            return result
        else:
            raise StopIteration

    def test_method(self) -> int:
        return 0

    @property
    def test_property(self) -> float:
        return self._private_attr

    @test_property.setter
    def test_property(self, value: Any) -> None:
        self._private_attr = value

    @staticmethod
    def static_method() -> bool:
        return True


class B(Enum):
    """This is a test enum to be used in functionality test."""

    Car = 1
    Cat = 2
    Dog = 3


@serializable(recursive_serde=True)
class IterWithoutLen:
    """This is a test class for testing iterator method in Klass."""

    __slots__ = ["n"]
    __attr_allowlist__ = ["n"]

    static_attr: int = 1

    def __iter__(self) -> "IterWithoutLen":
        self.n = 0
        return self

    def __next__(self) -> int:
        if self.n < A.static_attr:
            result = self.n
            self.n += 1
            return result
        else:
            raise StopIteration


@serializable(recursive_serde=True)
class C:

    __attr_allowlist__ = ["dynamic_object"]

    def __init__(self) -> None:
        self.dynamic_object = 123

    def dummy_reloadable_func(self) -> int:
        return 0

    @staticmethod
    def type_reload_func() -> None:
        C.dummy_reloadable_func = C.func_1  # type: ignore

    def obj_reload_func(self) -> None:
        print("RELOADING OBJ FUNCTION")
        self.dummy_reloadable_func = self.func_2  # type: ignore

    def func_1(self) -> int:
        return 1

    def func_2(self) -> int:
        return 2



"""A simple garbage collection heuristics.

Deleting a remote object using this logic will send a message
to the Data Owner indicating to remove the object from the store.
"""

# third party
from typing_extensions import final

# relative
from ...node.common.action.garbage_collect_object_action import (
    GarbageCollectObjectAction,
)
from ..pointer import Pointer
from .gc_strategy import GCStrategy


@final
class GCSimple(GCStrategy):
    """The GCSimple Strategy."""

    def reap(self, pointer: Pointer) -> None:
        """
        Send a message to delete the remote object.

        Args:
            pointer (Pointer): Pointer that indicates to an object
                        that should get deleted

        Return:
            None
        """

        # Create the delete message
        msg = GarbageCollectObjectAction(
            id_at_location=pointer.id_at_location, address=pointer.client.address
        )

        # Send the message
        pointer.client.send_eventual_msg_without_reply(msg=msg)



# stdlib
from collections import Counter
from collections import defaultdict
from typing import Any
from typing import Optional

# relative
from ....core.common.serde.serializable import serializable


@serializable(recursive_serde=True)
class DefaultDict(defaultdict):
    """Aim: build a class that inherits all the methods from collections.defaultdict but is serializable"""

    __attr_allowlist__ = []  # type: ignore

    def __init__(self, factory: Optional[Any] = None, **kwargs: Any) -> None:
        super().__init__(default_factory=factory, **kwargs)


@serializable(recursive_serde=True)
class SerializableCounter(Counter):

    __attr_allowlist__ = []  # type: ignore

    def __init__(self, iterable: Optional[Any] = None, **kwargs: Any) -> None:
        super().__init__(iterable, **kwargs)



# stdlib
from typing import Set

# relative
from ..location import Location


class LocationGroup(Location):

    """This represents a group of multiple locations. There
    are two kinds of LocationGroups, RegistryBackedLocationGroup
    and SubscriptionBackedLocationGroup. A registry backed group
    has an official owner with an official list while a subscription
    backed group is open to whoever knows the key to the group and
    chooses to subscribe. Both groups can have inclusion criteria
    which a location does or doesn't meet."""

    def __init__(self, known_group_members: Set[Location]):
        self.known_group_members = known_group_members
